(******************************************************************************
 * Mathias Bourgoin, Universit Pierre et Marie Curie (2012)
 *
 * Mathias.Bourgoin@gmail.com
 *
 * This software is a computer program whose purpose is to allow
 * GPU programming with the OCaml language.
 *
 * This software is governed by the CeCILL-B license under French law and
 * abiding by the rules of distribution of free software.  You can  use, 
 * modify and/ or redistribute the software under the terms of the CeCILL-B
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info". 
 * 
 * As a counterpart to the access to the source code and  rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty  and the software's author,  the holder of the
 * economic rights,  and the successive licensors  have only  limited
 * liability. 
 * 
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or 
 * data to be ensured and,  more generally, to use and operate it in the 
 * same conditions as regards security. 
 * 
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-B license and that you accept its terms.
*******************************************************************************)
open Spoc 
open Kernel
open Mem
open Devices
open Vector

open Kirc_Ast

type extension = 
	| ExFloat32
	| ExFloat64


type ('a) kirc_kernel =
{ 
  ml_kern : 'a;
  body : Kirc_Ast.k_ext;
  ret_val : Kirc_Ast.k_ext;
	extensions : extension array
}

let opencl_head = (
	"float fadd ( float a, float b ) { return (a + b);}\n"^
	"float fminus ( float a, float b ) { return (a - b);}\n"^
	"float fmul ( float a, float b ) { return (a * b);}\n"^
	"float fdiv ( float a, float b ) { return (a / b);}\n"
)
let opencl_float64 = (
	"#ifndef __FLOAT64_EXTENSION__ \n"^
	"#define __FLOAT64_EXTENSION__ \n"^
	"#if defined(cl_khr_fp64)  // Khronos extension available?\n"^
	"#pragma OPENCL EXTENSION cl_khr_fp64 : enable\n"^
	"#elif defined(cl_amd_fp64)  // AMD extension available?\n"^
	"#pragma OPENCL EXTENSION cl_amd_fp64 : enable\n"^
	"#endif\n"^
	"double dadd ( double a, double b ) { return (a + b);}\n"^
	"double dminus ( double a, double b ) { return (a - b);}\n"^
	"double dmul ( double a, double b ) { return (a * b);}\n"^
	"double ddiv ( double a, double b ) { return (a / b);}\n"^
	"#endif\n"
	)

let cuda_float64 = 
	(
	"#ifndef __FLOAT64_EXTENSION__ \n"^
	"#define __FLOAT64_EXTENSION__ \n"^
	"__device__ double dadd ( double a, double b ) { return (a + b);}\n"^
	"__device__ double dminus ( double a, double b ) { return (a - b);}\n"^
	"__device__ double dmul ( double a, double b ) { return (a * b);}\n"^
	"__device__ double ddiv ( double a, double b ) { return (a / b);}\n"^

	"#endif\n"
)
let cuda_head = (
	"__device__ float fadd ( float a, float b ) { return (a + b);}\n"^
	"__device__ float fminus ( float a, float b ) { return (a - b);}\n"^
	"__device__ float fmul ( float a, float b ) { return (a * b);}\n"^
	"__device__ float fdiv ( float a, float b ) { return (a / b);}\n"
	)
	
let new_var i = IdName ("spoc_var"^(string_of_int i))
let var i = IntId (("spoc_var"^(string_of_int i)), i)
let spoc_gen_kernel args body = Kern (args,body)
let spoc_fun_kernel a b = () 
let seq a b = Seq (a,b)
let app a b = App (a,b)
let spoc_unit () = Unit
let spoc_int a = Int a
let global_int_var a = GInt a
let global_float_var a = GFloat a
let spoc_int32 a = Int (Int32.to_int a)
let spoc_float f = Float f
let spoc_double d = Double d

let spoc_int_id a = Int a(*IntId (a,-1)*)
let spoc_float_id a = Float a
let spoc_plus a b = Plus (a,b)
let spoc_plus_float a b = Plusf (a,b)

let spoc_min a b = Min (a,b)
let spoc_min_float a b = Minf (a,b)

let spoc_mul a b = Mul (a,b)
let spoc_mul_float a b = Mulf (a,b)

let spoc_div a b = Div (a,b)
let spoc_div_float a b = Divf (a,b)

let spoc_mod a b = Mod (a,b)
let spoc_ife a b c = Ife (a,b,c)
let spoc_if a b  = If (a,b)
let spoc_do a b c d = DoLoop (a,b,c,d)
let spoc_while a b = While (a,b)

let params l = Params l
let spoc_id i = Id ("")
let spoc_return k = Return k 
let concat a b = Concat (a,b)
let empty_arg () = Empty
let new_int_var i = IntVar i 
let new_float_var i = FloatVar i 
let new_double_var i = DoubleVar i 
let new_unit_var i = UnitVar i

let new_int_vec_var v = VecVar (Int 0, v) 
let new_float_vec_var v = VecVar (Float 0., v) 
let new_double_vec_var v = VecVar (Double 0., v) 

let int_vect i = IntVect i 
let set_vect_var vecacc value = 
  SetV (vecacc, value)
let intrinsics a b = Intrinsics (a,b)  
let spoc_local_env local_var b = Local(local_var, b)
let spoc_set name value = Set (name, value)
let spoc_declare name = Decl (name)
let spoc_local_var a = a
let spoc_acc a b = Acc (a, b)
let int_var i = i
let int32_var i = i
let float_var f = f
let double_var d = CastDoubleVar d

let equals a b = EqBool (a,b)
let equals32 a b = EqBool (a,b)
let equals64 a b = EqBool (a,b)
let equalsF a b = EqBool (a,b)
let equalsF64 a b = EqBool (a,b)
let b_or a b = Or (a,b)
let b_and a b = And (a,b)

let lt a b = LtBool (a,b)
let lt32 a b = LtBool (a,b)
let lt64 a b = LtBool (a,b)
let ltF a b = LtBool (a,b)
let ltF64 a b = LtBool (a,b)

let gt a b = GtBool (a,b)
let gt32 a b = GtBool (a,b)
let gt64 a b = GtBool (a,b)
let gtF a b = GtBool (a,b)
let gtF64 a b = GtBool (a,b)

let lte a b = LtEBool (a,b)
let lte32 a b = LtEBool (a,b)
let lte64 a b = LtEBool (a,b)
let lteF a b = LtEBool (a,b)
let lteF64 a b = LtEBool (a,b)

let gte a b = GtEBool (a,b)
let gte32 a b = GtEBool (a,b)
let gte64 a b = GtEBool (a,b)
let gteF a b = GtEBool (a,b)
let gteF64 a b = GtEBool (a,b)

let get_vec a b = IntVecAcc (a,b)
let return_unit () =  Unit

let return_int i = IntVar i
let return_float f = FloatVar f
let return_double d = DoubleVar d





let print s = Printf.printf "%s}\n" s 

let debug_print ((kir : ('a, 'b) spoc_kernel),
		 (k: ('c)  kirc_kernel)) =
  let (k1,k2,k3) = (k.ml_kern, k.body,k.ret_val) in 
  print_ast k2
;;


let a_to_vect = function
  | IntVar i  ->  (new_int_vec_var ( i))
  | a  -> print_ast a; failwith "a_to_vect"

let a_to_return_vect k1 k2 idx= 
  match k1 with
  | IntVar i  ->  (set_vect_var (get_vec (var i) idx) (k2))
  | _  -> failwith "error a_to_return_vect"

let param_list = ref []


let add_to_param_list a = 
  param_list := a :: !param_list
    
let rec check_and_transform_to_map a = 
  match a with 
  | Plus (b,c)  -> Plus(check_and_transform_to_map b, check_and_transform_to_map c)
  | Min (b,c)  -> Min(check_and_transform_to_map b, check_and_transform_to_map c)
  | Mul (b,c)  -> Mul(check_and_transform_to_map b, check_and_transform_to_map c)
  | Mod (b,c)  -> Mod(check_and_transform_to_map b, check_and_transform_to_map c)
  | Div (b,c)  -> Div(check_and_transform_to_map b, check_and_transform_to_map c)
  | IntId (v,i)  -> 
    if (List.mem i !param_list) then
      IntVecAcc(IdName ("spoc_var"^(string_of_int i)), 
		Intrinsics ("blockIdx.x*blockDim.x+threadIdx.x","get_global_id (0)"))
    (*(IntId ("spoc_global_id", -1)))*)
    else 
      a
  | _  -> a
    
let map (k: ('c) kirc_kernel) (ker2: ('a, 'b) spoc_kernel)= 
  let (k1,k2,k3) = (k.ml_kern, k.body,k.ret_val) in 
  param_list := [];
  let rec aux = function
    | Kern (args, body)  ->  
      let new_args = 
        match args with
        | Params p ->
          (match p with 
          | Concat (Concat _, _) ->
	    failwith "error multiple map args ";  
	  | Concat (a, Empty)  ->
            params (concat (a_to_vect a) (concat (a_to_vect k3) (empty_arg ()))) 
          | _ -> failwith "map type error")
        | _  -> failwith "error map args"
      in let n_body =
	   let rec aux2  curr =
	     match curr with
	     | Return _ 
	     | Local _ 
	     | Seq _ -> aux curr
	     | Plus (a,b) -> a_to_return_vect k3 (aux curr) ((intrinsics "blockIdx.x*blockDim.x+threadIdx.x" "get_global_id(0)"))
	     | _ -> Printf.printf "-----> "; print_ast curr; failwith "map aux2"		 
	   and aux curr =
             match curr with
             | Return a  -> aux2 a
             | Seq (a,b)  -> seq a (aux2 b)
             | Local (a,b) -> Local (a, aux2 b)
             | Plus (a,b)  -> 
	       Plus (aux a, aux b)
	     | Min (a,b)  -> Min (a,b)
	     | Mul (a,b)  -> Mul (a,b)
	     | Div (a,b)  -> Div (a,b)
	     | Mod (a,b)  -> Mod (a,b)
	     | IntId (v,i)  -> 
	       if i = 0 then
		 IntVecAcc(IdName ("spoc_var"^(string_of_int i)), 
			   Intrinsics ("blockIdx.x*blockDim.x+threadIdx.x","get_global_id (0)"))
	       else
		 curr 
	     | a -> a
           in
           aux body
	 in
         Kern (new_args, n_body)
    | _ -> failwith "malformed kernel for map"   
  in 
  (ker2,
   { 
     ml_kern = Tools.map (k1);
     body = aux k2;
     ret_val = Unit;
		 extensions = k.extensions;
   })
  

let rewrite ker =
  let b = ref false in 
  let rec aux kern = 
    match kern with
    | Kern (k1,k2) -> 
      Kern (aux k1, aux k2)
    | Params k -> 
      Params (aux k)
    | Plus (k1,k2) -> 
      Plus (aux k1, aux k2)
    | Plusf (k1,k2) -> 
      (match k1,k2 with
      | Float f1, Float f2 -> 
	(b := true; Float (f1 +. f2))
      |  _ -> Plusf (aux k1, aux k2))
    | Min (k1,k2) -> 
      Min (aux k1, aux k2)
    | Minf (k1,k2) -> 
      (match k1,k2 with
      | Float f1, Float f2 -> 
	(b := true; Float (f1 +. f2))
      |  _ -> Minf (aux k1, aux k2))
    | Mul (k1,k2) -> 
      Mul (aux k1, aux k2)
    | Mulf (k1,k2) -> 
      (match k1,k2 with
      | Float f1, Float f2 -> 
	(b := true; Float (f1 +. f2))
      |  _ -> Mulf (aux k1, aux k2))
    | Div (k1,k2) -> 
      Div (aux k1, aux k2)
    | Divf (k1,k2) -> 
      (match k1,k2 with
      | Float f1, Float f2 -> 
	(b := true; Float (f1 +. f2))
      |  _ -> Divf (aux k1, aux k2))
    | Mod (k1,k2) -> 
      Mod (aux k1, aux k2)
    | Id _ -> kern
    | IdName _ -> kern
    | IntVar _ -> kern
    | FloatVar _ -> kern
    | UnitVar _ ->
      Seq (kern, kern)
    | CastDoubleVar _ -> kern
    | DoubleVar _ -> kern
    | VecVar (k,idx ) ->
      VecVar (aux k, idx) 
    | Concat (k1,k2) -> 
      Concat (aux k1, aux k2)
    | Empty -> kern
    | Seq (k1,k2) -> 
      Seq (aux k1, aux k2)
    | Return k -> 
      (match k with
      | Ife(k1, k2, k3) ->
	(b := true;
	 Ife (aux k1, aux (Return k2), aux (Return k3)))
      | If(k1, k2) ->
	(b := true;
	 If (aux k1, aux (Return k2)))
      | DoLoop (k1,k2,k3,k4) ->
	(b := true;
	 DoLoop (aux k1, aux k2, aux k3, aux (Return k4)))
      | While (k1, k2) ->
	(b:= true;
	 While (aux k1, aux (Return k2)))
      | Seq (k1, k2) ->
	( b:= true;
	  Seq ( aux k1, aux (Return k2)))
      | _ ->
	Return (aux k))
    | Acc (k1,k2) ->
      Acc (aux k1, aux k2)
    | Set (k1,k2) ->
      Set (aux k1, aux k2)
    | Decl k1 -> aux k1
    | SetV (k1,k2) -> 
      (match k2 with
      | Seq (k3, k4) ->
	Seq (k3, SetV (aux k1, aux k4))
      | _ -> SetV (aux k1, aux k2)
      )
    | SetLocalVar (k1,k2,k3) -> 
      SetLocalVar (aux k1, aux k2, aux k3) 
    | Intrinsics _ -> kern
    | IntId _ -> kern
    | Int _-> kern
    | GInt _ -> kern
    | GFloat _ -> kern
    | Float _ -> kern
    | Double _ -> kern
    | IntVecAcc (k1,k2) -> 
      (match k2 with
      |	Seq (k3,k4) ->
	Seq(k3, IntVecAcc (aux k1, aux k4))
      | _ -> IntVecAcc (aux k1, aux k2))
    | Local (k1,k2) -> 
      Local (aux k1, aux k2)
    | Ife (k1,k2,k3) ->
      Ife (aux k1, aux k2, aux k3)   
    | If (k1,k2) ->
      If (aux k1, aux k2)   
    | Or (k1,k2) -> 
      Or (aux k1, aux k2)
    | And (k1,k2) -> 
      And (aux k1, aux k2)
    | EqBool (k1,k2) -> 
      EqBool (aux k1, aux k2)
    | LtBool (k1,k2) -> 
      LtBool (aux k1, aux k2)
    | GtBool (k1,k2) -> 
      GtBool (aux k1, aux k2)
    | LtEBool (k1,k2) -> 
      LtEBool (aux k1, aux k2)
    | GtEBool (k1,k2) -> 
      GtEBool (aux k1, aux k2)
    | DoLoop (k1, k2, k3, k4) ->
      DoLoop (aux k1, aux k2, aux k3, aux k4)
    | While (k1, k2) ->
      While (aux k1, aux k2)
    | App (a,b) -> App (aux a, (Array.map aux b))
    | Unit -> kern
      
  in
  let kern = ref (aux ker) in
  while (!b) do
    b := false;
    kern := aux !kern;
  done;
  !kern
;;



let return_v = ref ("","")

let save file string =
  let channel = open_out file in
  output_string channel string;
  close_out channel;;

let load_file f =
  let ic = open_in f in
  let n = in_channel_length ic in
  let s = String.create n in
  really_input ic s 0 n;
  close_in ic;
  (s)

let gen ((kir: ('a, 'b) spoc_kernel),(k: ('c) kirc_kernel)) = 
  let (k1,k2,k3) = (k.ml_kern, k.body,k.ret_val) in
  return_v := "","";
  let k' = ((Kirc_Cuda.parse 0 k3),
	   ( match  k3 with
	   | IntVar i | FloatVar i | DoubleVar i -> "spoc_var"^(string_of_int i)^" = "
	   | Unit -> ""
	   | SetV _ -> ""
	   | IntVecAcc _-> ""
	   | VecVar _ -> ""
	   | _ -> (debug_print 
		     (kir,
		      {ml_kern = k1;
		       body = k3;
		       ret_val = k3;
					 extensions = k.extensions});  Pervasives.flush stdout; assert false) ))
  in
  
  Kirc_Cuda.return_v := k';
  Kirc_OpenCL.return_v := k';
	let cuda_head = 
		Array.fold_left 
			(fun header extension -> 
				match extension with
				| ExFloat32 -> header
				| ExFloat64 -> cuda_float64^header) cuda_head k.extensions in
	let opencl_head = 
		Array.fold_left 
			(fun header extension -> 
				match extension with
				| ExFloat32 -> header
				| ExFloat64 -> opencl_float64^header) opencl_head k.extensions in		
  
	kir#set_opencl_sources (opencl_head^(Kirc_OpenCL.parse 0 (rewrite k2)));
	
	save "kirc_kernel.cu" (cuda_head^(Kirc_Cuda.parse 0 (rewrite k2))) ;
(*  save "kirc_kernel.cl" (List.hd (kir#get_opencl_sources ())); *)
	ignore(Sys.command ("nvcc -O3 -g -ptx kirc_kernel.cu -o kirc_kernel.ptx"));
  let s = (load_file "kirc_kernel.ptx") in
	kir#set_cuda_sources s;
  ignore(Sys.command "rm kirc_kernel.cu kirc_kernel.ptx"); 

  kir#reset_binaries ();
  kir,k
    
 
let run ((kir: ('a, 'b) spoc_kernel),(k: ('c) kirc_kernel)) = 
  kir#run
    

let compile_kernel_to_files s ((kir: ('a, 'b) spoc_kernel),(k: ('c) kirc_kernel)) =
	 let (k1,k2,k3) = (k.ml_kern, k.body,k.ret_val) in
  return_v := "","";
	 let k' = ((Kirc_Cuda.parse 0 k3),
	   ( match  k3 with
	   | IntVar i | FloatVar i | DoubleVar i -> "spoc_var"^(string_of_int i)^" = "
	   | Unit -> ""
	   | SetV _ -> ""
	   | IntVecAcc _-> ""
	   | VecVar _ -> ""
	   | _ -> (debug_print 
		     (kir,
		      {ml_kern = k1;
		       body = k3;
		       ret_val = k3;
					 extensions = k.extensions});  Pervasives.flush stdout; assert false) ))
  in
  
  Kirc_Cuda.return_v := k';
  Kirc_OpenCL.return_v := k';
	let cuda_head = 
		Array.fold_left 
			(fun header extension -> 
				match extension with
				| ExFloat32 -> header
				| ExFloat64 -> cuda_float64^header) cuda_head k.extensions in
	let opencl_head = 
		Array.fold_left 
			(fun header extension -> 
				match extension with
				| ExFloat32 -> header
				| ExFloat64 -> opencl_float64^header) opencl_head k.extensions in			
	save (s^".cu") (cuda_head^(Kirc_Cuda.parse 0 (rewrite k2))) ;
  save (s^".cl") (opencl_head^(Kirc_OpenCL.parse 0 (rewrite k2)))

	

module Std =
struct
  let thread_idx_x = 1
  let thread_idx_y = 1
  let thread_idx_z = 1
  let block_idx_x = 1
  let block_idx_y = 1
  let block_idx_z = 1
  let block_dim_x = 1
  let block_dim_y = 1
  let block_dim_z = 1

  let global_thread_id = 0
  let return () = ()
	
	let float64 = float
	let int_of_float64 = int_of_float
	
end

module Math =
struct
	module Float32 =
	struct
		let add = (+.)
		let minus = (-.)
		let mul = ( *. )
		let div = (/.)
		
		let pow = ( ** )	
		let sqrt = Pervasives.sqrt
		let exp = Pervasives.exp
		let log = Pervasives.log
		let log10 = Pervasives.log10
		let expm1 = Pervasives.expm1
		let log1p = Pervasives.log1p
		
		let acos = Pervasives.acos
		let cos = Pervasives.cos
		let cosh = Pervasives.cosh
		let asin = Pervasives.asin
		let sin = Pervasives.sin
		let sinh = Pervasives.sinh
		let tan = Pervasives.tan
		let tanh = Pervasives.tanh
		let atan = Pervasives.atan
		let atan2 = Pervasives.atan2
		let hypot = Pervasives.hypot
		
		let ceil = Pervasives.ceil
		let floor = Pervasives.floor
		
		let abs_float = Pervasives.abs_float
		let copysign = Pervasives.copysign
		let modf = Pervasives.modf		
		
		let zero = 0.
		let one = 1.
		let of_float (f:float) = f
		let to_float (f:float) = f

	end
	
	module Float64 =
	struct
		let add = (+.)
		let minus = (-.)
		let mul = ( *. )
		let div = (/.)
		
		let pow = ( ** )	
		let sqrt = Pervasives.sqrt
		let exp = Pervasives.exp
		let log = Pervasives.log
		let log10 = Pervasives.log10
		let expm1 = Pervasives.expm1
		let log1p = Pervasives.log1p
		
		let acos = Pervasives.acos
		let cos = Pervasives.cos
		let cosh = Pervasives.cosh
		let asin = Pervasives.asin
		let sin = Pervasives.sin
		let sinh = Pervasives.sinh
		let tan = Pervasives.tan
		let tanh = Pervasives.tanh
		let atan = Pervasives.atan
		let atan2 = Pervasives.atan2
		let hypot = Pervasives.hypot
		
		let ceil = Pervasives.ceil
		let floor = Pervasives.floor
		
		let abs_float = Pervasives.abs_float
		let copysign = Pervasives.copysign
		let modf = Pervasives.modf	
		
		let zero = 0.
		let one = 1.
		let of_float (f:float) = f
		let to_float (f:float) = f
	end
end
