(******************************************************************************
 * Mathias Bourgoin, Université Pierre et Marie Curie (2012)
 *
 * Mathias.Bourgoin@gmail.com
 *
 * This software is a computer program whose purpose is to allow GPU 
 * programming with the OCaml language.
 *
 * This software is governed by the CeCILL-B license under French law and
 * abiding by the rules of distribution of free software.  You can  use, 
 * modify and/ or redistribute the software under the terms of the CeCILL-B
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info". 
 * 
 * As a counterpart to the access to the source code and  rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty  and the software's author,  the holder of the
 * economic rights,  and the successive licensors  have only  limited
 * liability. 
 * 
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or 
 * data to be ensured and,  more generally, to use and operate it in the 
 * same conditions as regards security. 
 * 
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-B license and that you accept its terms.
 ******************************************************************************)

open Camlp4.PreCast
open Syntax
open Ast

let my_eprintf s = output_string stderr s
let my_eprintf a = () 

type ktyp =
| TUnknown
| TUnit
| TInt
| TInt32
| TInt64
| TFloat
| TFloat32
| TFloat64
| TBool
| TVec of ktyp
| TApp of ktyp * ktyp
    
let rec ktyp_to_string = function
  | TUnit -> "unit"
  | TInt -> "int"
  | TInt32  -> "int32"
  | TInt64  -> "int64"
  | TFloat  ->  "float"
  | TFloat32 -> "float32"
  | TFloat64 -> "float64"
  | TUnknown  -> "unknown"
  | TBool -> "bool"
  | TVec k -> (ktyp_to_string k)^" vector" 
  | TApp (k1,k2) -> (ktyp_to_string k1)^" -> "^(ktyp_to_string k2) 
 (* | _ ->  assert false  *)




type k_expr = 
| Open of Loc.t * ident * kexpr
| App of Loc.t * kexpr * kexpr list
| Acc of Loc.t * kexpr * kexpr
| VecSet of Loc.t*kexpr*kexpr
| VecGet of Loc.t*kexpr*kexpr
| Seq of Loc.t*kexpr*kexpr
| Seqs of kexpr list
| Bind of Loc.t*kexpr*kexpr*kexpr*bool
| Plus of Loc.t*kexpr*kexpr
| Plus32 of Loc.t*kexpr*kexpr
| Plus64 of Loc.t*kexpr*kexpr
| PlusF of Loc.t*kexpr*kexpr
| PlusF32 of Loc.t*kexpr*kexpr
| PlusF64 of Loc.t*kexpr*kexpr

| Min of Loc.t*kexpr*kexpr
| Min32 of Loc.t*kexpr*kexpr
| Min64 of Loc.t*kexpr*kexpr
| MinF of Loc.t*kexpr*kexpr
| MinF32 of Loc.t*kexpr*kexpr
| MinF64 of Loc.t*kexpr*kexpr

| Mul of Loc.t*kexpr*kexpr
| Mul32 of Loc.t*kexpr*kexpr
| Mul64 of Loc.t*kexpr*kexpr
| MulF of Loc.t*kexpr*kexpr
| MulF32 of Loc.t*kexpr*kexpr
| MulF64 of Loc.t*kexpr*kexpr

| Mod of Loc.t*kexpr*kexpr

| Div of Loc.t*kexpr*kexpr
| Div32 of Loc.t*kexpr*kexpr
| Div64 of Loc.t*kexpr*kexpr
| DivF of Loc.t*kexpr*kexpr
| DivF32 of Loc.t*kexpr*kexpr
| DivF64 of Loc.t*kexpr*kexpr

| Id of Loc.t*ident
| Int of Loc.t*string
| Int32 of Loc.t*string
| Int64 of Loc.t*string
| Float of Loc.t*string
| Float32 of Loc.t*string
| Float64 of Loc.t*string
| CastId of ktyp* k_expr
| BoolAnd of Loc.t*kexpr*kexpr
| BoolOr of Loc.t*kexpr*kexpr
| BoolEq of Loc.t*kexpr*kexpr
| BoolEq32 of Loc.t*kexpr*kexpr
| BoolEq64 of Loc.t*kexpr*kexpr
| BoolEqF of Loc.t*kexpr*kexpr
| BoolEqF64 of Loc.t*kexpr*kexpr
| BoolLt of Loc.t*kexpr*kexpr
| BoolLt32 of Loc.t*kexpr*kexpr
| BoolLt64 of Loc.t*kexpr*kexpr
| BoolLtF of Loc.t*kexpr*kexpr
| BoolLtF64 of Loc.t*kexpr*kexpr

| BoolLtE of Loc.t*kexpr*kexpr
| BoolLtE32 of Loc.t*kexpr*kexpr
| BoolLtE64 of Loc.t*kexpr*kexpr
| BoolLtEF of Loc.t*kexpr*kexpr
| BoolLtEF64 of Loc.t*kexpr*kexpr

| BoolGt of Loc.t*kexpr*kexpr
| BoolGt32 of Loc.t*kexpr*kexpr
| BoolGt64 of Loc.t*kexpr*kexpr
| BoolGtF of Loc.t*kexpr*kexpr
| BoolGtF64 of Loc.t*kexpr*kexpr

| BoolGtE of Loc.t*kexpr*kexpr
| BoolGtE32 of Loc.t*kexpr*kexpr
| BoolGtE64 of Loc.t*kexpr*kexpr
| BoolGtEF of Loc.t*kexpr*kexpr
| BoolGtEF64 of Loc.t*kexpr*kexpr

| Ife of Loc.t * kexpr * kexpr * kexpr
| If of Loc.t * kexpr * kexpr 
| DoLoop of Loc.t * kexpr * kexpr * kexpr * kexpr
| While of Loc.t * kexpr * kexpr
| End of Loc.t*kexpr
| Ref of Loc.t*kexpr

| ModuleAccess of Loc.t * string * kexpr

| Noop

and kexpr = {
  mutable t : ktyp;
  mutable e: k_expr;
  loc: Loc.t}

let rec k_expr_to_string = function
  | App _ -> "App"
  | Acc _ -> "Acc"
  | VecSet _ -> "VecSet"
  | VecGet _ -> "VecGet"
  | Seq _ -> "Seq"
  | Bind _ -> "Bind"

  | Plus _ -> "Plus"
  | Plus32 _ -> "Plus32"
  | Plus64 _ -> "Plus64"
  | PlusF _ -> "PlusF"
  | PlusF32 _ -> "PlusF32"
  | PlusF64 _ -> "PlusF64"
  | Min _ -> "Min"
  | Min32 _ -> "Min32"
  | Min64 _ -> "Min64"
  | MinF _ -> "MinF"
  | MinF32 _ -> "MinF32"
  | MinF64 _ -> "MinF64"
  | Mul _ -> "Mul"
  | Mul32 _ -> "Mul32"
  | Mul64 _ -> "Mul64"
  | MulF _ -> "MulF"
  | MulF32 _ -> "MulF32"
  | MulF64 _ -> "MulF64"
  | Div _ -> "Div"
  | Div32 _ -> "Div32"
  | Div64 _ -> "Div64"
  | DivF _ -> "DivF"
  | DivF32 _ -> "DivF32"
  | DivF64 _ -> "DivF64"
  | Mod _ -> "Mod"
  | Id _ -> "Id"
  | Int _ -> "Int"
  | Int32 _ -> "Int32"
  | Int64 _ -> "Int64"
  | Float _ -> "Float"
  | Float32 _ -> "Float32"
  | Float64 _ -> "Float64"
  | CastId _ -> "CastId"
  | BoolEq _ -> "BoolEq"
  | BoolEq32 _ -> "BoolEq32"
  | BoolEq64 _ -> "BoolEq64"
  | BoolEqF _ -> "BoolEqF"
  | BoolEqF64 _ -> "BoolEqF64"
  | BoolLt _ -> "BoolEq"
  | BoolLt32 _ -> "BoolEq32"
  | BoolLt64 _ -> "BoolEq64"
  | BoolLtF _ -> "BoolEqF"
  | BoolLtF64 _ -> "BoolEqF64"
  | BoolGt _ -> "BoolGt"
  | BoolGt32 _ -> "BoolGt32"
  | BoolGt64 _ -> "BoolGt64"
  | BoolGtF _ -> "BoolGtF"
  | BoolGtF64 _ -> "BoolGtF64"
  | BoolLtE _ -> "BoolLtE"
  | BoolLtE32 _ -> "BoolLtE32"
  | BoolLtE64 _ -> "BoolLtE64"
  | BoolLtEF _ -> "BoolLtEF"
  | BoolLtEF64 _ -> "BoolLtEF64"
  | BoolGtE _ -> "BoolGtE"
  | BoolGtE32 _ -> "BoolGtE32"
  | BoolGtE64 _ -> "BoolGtE64"
  | BoolGtEF _ -> "BoolGtEF"
  | BoolGtEF64 _ -> "BoolGtEF64"

  | BoolAnd _ -> "BoolAnd"
  | BoolOr _ -> "BoolOr"
  | Ife _ -> "Ife"
  | If _ -> "If"
  | DoLoop _ -> "DoLoop"
  | While _ -> "While"
  | End _ -> "End"
  | Seqs _ -> "Seqs"
  | Open _ -> "Open"
  | Noop -> "Noop"
  | ModuleAccess _ -> "ModuleAccess"
  | Ref _ -> "Ref"


let ex32 = 
	let _loc = Loc.ghost in
	<:expr< ExFloat32>>
let ex64 = 	let _loc = Loc.ghost in
	<:expr< ExFloat64>>

let extensions = 
		ref [ ex32 ] 
	  
type var = {
  n : int;
  mutable var_type: ktyp;
  mutable is_mutable : bool;
  mutable read_only : bool;
  mutable write_only : bool;
  mutable is_global : bool;
}

let new_kernel = ref false

exception ArgumentError
exception Unbound_value of string * Loc.t
exception Unbound_module of string * Loc.t
exception TypeError of ktyp * ktyp * Loc.t

let rec string_of_ident i = 
  let aux = function
    | <:ident< $lid:s$ >> -> s
    | <:ident< $uid:s$ >> -> s
    | <:ident< $i1$.$i2$ >> -> "" ^ (string_of_ident i1) ^ "." ^ (string_of_ident i2)
    | <:ident< $i1$($i2$) >> -> "" ^ (string_of_ident i1) ^ " " ^ (string_of_ident i2)
		| _ -> assert false
  in aux i

let expr_of_patt p =
  (*let _loc = Loc.ghost in*)
  match p with
	| PaId (l,i) -> ExId (l,i)
(*    <:patt< $lid:x$ >> -> <:expr< $lid:x$ >> *)
  | _ -> raise ArgumentError

type k_patt 


type cfun = 
{ nb_args: int;
  cuda_val : string;
  opencl_val : string;
  typ : ktyp 
}
  
type spoc_module = {
       mod_name : string;
       mod_constants : (ktyp * string * string * string) list;
       mod_functions : (ktyp * string * int * string * string) list;
       mod_modules : (string, spoc_module) Hashtbl.t
} 


let return_type = ref TUnknown

let arg_idx = ref 0

let args () = 
  let tbl = Hashtbl.create 10 in
  tbl
let current_args = ref (args ()) 

let intrinsics_fun = ref (Hashtbl.create 100)
let intrinsics_const = ref (Hashtbl.create 100)

let arg_list = ref []


  
let std = {
  mod_name = "Std";
  mod_constants = 
    [
      (TInt, "thread_idx_x", "threadIdx.x", "(get_local_id (0))");
      (TInt, "thread_idx_y", "threadIdx.y", "(get_local_id (1))");
      (TInt, "thread_idx_z", "threadIdx.z", "(get_local_id (2))");
      
      (TInt, "block_idx_x", "blockIdx.x", "(get_group_id (0))");
      (TInt, "block_idx_y", "blockIdx.y", "(get_group_id (1))");
      (TInt, "block_idx_z", "blockIdx.z", "(get_group_id (2))");

      (TInt, "block_dim_x", "blockDim.x", "(get_local_size (0))");
      (TInt, "block_dim_y", "blockDim.y", "(get_local_size (1))");
      (TInt, "block_dim_z", "blockDim.z", "(get_local_size (2))");

      (TInt, "grid_dim_x", "gridDim.x", "(get_num_groups (0))");
      (TInt, "grid_dim_y", "gridDim.y", "(get_num_groups (1))");
      (TInt, "grid_dim_z", "gridDim.z", "(get_num_groups (2))");

      (TInt, "global_thread_id", "blockIdx.x*blockDim.x+threadIdx.x", 
       "get_global_id (0)");
			
			 (TFloat64, "zero64", "0.", "0.");
			 (TFloat64, "one64", "1.", "1.");
    ];
  mod_functions = [
    (TApp (TUnit, TUnit), "return", 1, "return", "return");

    (TApp (TInt, TFloat32), "float_of_int", 1, "(float)", "(float)");
    (TApp (TInt, TFloat32), "float", 1, "(float)", "(float)");

    (TApp (TInt, TFloat64), "float64_of_int", 1, "(double)", "(double)");
    (TApp (TInt, TFloat64), "float64", 1, "(double)", "(double)");
		(TApp (TFloat, TFloat64), "float64_of_float", 1, "(double)", "(double)");

    (TApp (TFloat32, TInt), "int_of_float", 1, "(int)", "(int)");
    (TApp (TFloat64, TInt), "int_of_float64", 1, "(int)", "(int)");

  ];
  mod_modules = 
		let m = Hashtbl.create 2 in 
		m
}

let mathf32 = {
  mod_name = "Float32";
  mod_constants = 
    [
			(TFloat32, "zero", "0.f", "0.f");
			(TFloat32, "one", "1.f", "1.f");
    ];
  mod_functions = [
		(TApp ((TApp (TFloat32, TFloat32)), TFloat32), "add", 2, "fadd", "fadd");
		(TApp ((TApp (TFloat32, TFloat32)), TFloat32), "minus", 2, "fminus", "fminus");
		(TApp ((TApp (TFloat32, TFloat32)), TFloat32), "mul", 2, "fmul", "fmul");
		(TApp ((TApp (TFloat32, TFloat32)), TFloat32), "div", 2, "fdiv", "fdiv");

		(TApp ((TApp (TFloat32, TFloat32)), TFloat32), "pow", 2, "powf", "pow");
		(TApp (TFloat32, TFloat32), "sqrt", 1, "sqrtf", "sqrt");
		(TApp (TFloat32, TFloat32), "exp", 1, "expf", "exp");
		(TApp (TFloat32, TFloat32), "log", 1, "logf", "log");
		(TApp (TFloat32, TFloat32), "log10", 1, "log10f", "log10");
		(TApp (TFloat32, TFloat32), "expm1", 1, "expm1f", "expm1");
		(TApp (TFloat32, TFloat32), "log1p", 1, "log1pf", "log1p");


    (TApp (TFloat32, TFloat32), "acos", 1, "acosf", "acos");
		(TApp (TFloat32, TFloat32), "cos", 1, "cosf", "cos");
		(TApp (TFloat32, TFloat32), "cosh", 1, "coshf", "cosh");
    (TApp (TFloat32, TFloat32), "asin", 1, "asinf", "asin");
    (TApp (TFloat32, TFloat32), "sin", 1, "sinf", "sin");
    (TApp (TFloat32, TFloat32), "sinh", 1, "sinhf", "sinh");
 		(TApp (TFloat32, TFloat32), "tan", 1, "tanf", "tan");
 		(TApp (TFloat32, TFloat32), "tanh", 1, "tanhf", "tanh");
		(TApp (TFloat32, TFloat32), "atan", 1, "atanf", "atan");
 		(TApp ((TApp (TFloat32, TFloat32)), TFloat32), "atan2", 2, "atan2f", "atan2");
		(TApp ((TApp (TFloat32, TFloat32)), TFloat32), "hypot", 2, "hypotf", "hypot");

		
 		(TApp (TFloat32, TFloat32), "ceil", 1, "ceilf", "ceil");
 		(TApp (TFloat32, TFloat32), "floor", 1, "floorf", "floor");
 		
		(TApp (TFloat32, TFloat32), "abs_float", 1, "fabsf", "fabs");

		(TApp ((TApp (TFloat32, TFloat32)), TFloat32), "copysign", 2, "copysignf", "copysign");
		(TApp ((TApp (TFloat32, TFloat32)), TFloat32), "modf", 2, "fmodf", "fmod");
		
		(TApp (TFloat, TFloat32), "of_float", 1, "(float)", "(float)");
		(TApp (TFloat32, TFloat), "to_float", 1, "(float)", "(float)");


  ];
  mod_modules = Hashtbl.create 0
}

let mathf64 = {
  mod_name = "Float64";
  mod_constants = 
    [
			(TFloat64, "zero", "0.", "0.");
			(TFloat64, "one", "1.", "1.");
    ];
  mod_functions = [
    (TApp ((TApp (TFloat64, TFloat64)), TFloat64), "add", 2, "dadd", "dadd");
		(TApp ((TApp (TFloat64, TFloat64)), TFloat64), "minus", 2, "dminus", "dminus");
		(TApp ((TApp (TFloat64, TFloat64)), TFloat64), "mul", 2, "dmul", "dmul");
		(TApp ((TApp (TFloat64, TFloat64)), TFloat64), "div", 2, "ddiv", "ddiv");

		(TApp ((TApp (TFloat64, TFloat64)), TFloat64), "pow", 2, "powf", "pow");
		(TApp (TFloat64, TFloat64), "sqrt", 1, "sqrt", "sqrt");
		(TApp (TFloat64, TFloat64), "exp", 1, "exp", "exp");
		(TApp (TFloat64, TFloat64), "log", 1, "log", "log");
		(TApp (TFloat64, TFloat64), "log10", 1, "log10", "log10");
		(TApp (TFloat64, TFloat64), "expm1", 1, "expm1", "expm1");
		(TApp (TFloat64, TFloat64), "log1p", 1, "log1p", "log1p");


    (TApp (TFloat64, TFloat64), "acos", 1, "acos", "acos");
		(TApp (TFloat64, TFloat64), "cos", 1, "cos", "cos");
		(TApp (TFloat64, TFloat64), "cosh", 1, "cosh", "cosh");
    (TApp (TFloat64, TFloat64), "asin", 1, "asin", "asin");
    (TApp (TFloat64, TFloat64), "sin", 1, "sin", "sin");
    (TApp (TFloat64, TFloat64), "sinh", 1, "sinh", "sinh");
 		(TApp (TFloat64, TFloat64), "tan", 1, "tan", "tan");
 		(TApp (TFloat64, TFloat64), "tanh", 1, "tanh", "tanh");
		(TApp (TFloat64, TFloat64), "atan", 1, "atan", "atan");
 		(TApp ((TApp (TFloat64, TFloat64)), TFloat64), "atan2", 2, "atan2", "atan2");
		(TApp ((TApp (TFloat64, TFloat64)), TFloat64), "hypot", 2, "hypot", "hypot");

		
 		(TApp (TFloat64, TFloat64), "ceil", 1, "ceil", "ceil");
 		(TApp (TFloat64, TFloat64), "floor", 1, "floor", "floor");
 		
		(TApp (TFloat64, TFloat64), "abs_float", 1, "fabs", "fabs");

		(TApp ((TApp (TFloat64, TFloat64)), TFloat64), "copysign", 2, "copysign", "copysign");
		(TApp ((TApp (TFloat64, TFloat64)), TFloat64), "modf", 2, "fmod", "fmod");

		(TApp (TFloat, TFloat64), "of_float", 1, "(double)", "(double)");
		(TApp (TFloat64, TFloat), "to_float", 1, "(double)", "(double)");


  ];
  mod_modules = Hashtbl.create 0
}


let math = {
  mod_name = "Math";
  mod_constants = 
    [
    ];
  mod_functions = [
  ];
  mod_modules = 
    let m = Hashtbl.create 0 in
    Hashtbl.add m mathf32.mod_name mathf32; 
    Hashtbl.add m mathf64.mod_name mathf64;
    m
}


let modules = 
  let m = Hashtbl.create 10 in
  Hashtbl.add m std.mod_name std;
  Hashtbl.add m math.mod_name math; 
  m

	 
(*let new_var var i  value typ =
  (*incr arg_idx; *)
  match var.e with 
  | Id (_loc, s)  -> 
    let ex1,ex2 = 
      match typ with
      | TInt -> var.t <- TInt; <:expr<(new_int_var $ExInt(_loc,string_of_int i)$)>>, <:expr<( $value$)>>
      | TInt32 -> var.t <- TInt32; <:expr<(new_int32_var $ExInt(_loc,string_of_int i)$)>>, <:expr<( $value$)>>
      | TInt64 -> var.t <- TInt64; <:expr<(new_int64_var $ExInt(_loc,string_of_int i)$)>>, <:expr<( $value$)>>
      | TFloat -> var.t <- TFloat; <:expr<(new_float_var $ExInt(_loc,string_of_int i)$)>>,<:expr<( $value$)>>
      | TFloat64 -> var.t <- TFloat; <:expr<(new_double_var $ExInt(_loc,string_of_int i)$)>>,<:expr<( $value$)>>
      | _  -> (* assert false *) (); failwith "unknown var type"
    in
    incr arg_idx;
    Hashtbl.add !current_args (string_of_ident s) {n=i; var_type=var.t};
    arg_list := <:expr<(spoc_declare $ex1$) >>:: !arg_list
		(*<:expr<new_var $ExInt(_loc,string_of_int i)$>>*)
  | CastId (tt,Id(_loc, s))  -> 
    let ex1,ex2 = 
      match typ with
      | TInt -> var.t <- TInt; <:expr<(new_int_var $ExInt(_loc,string_of_int i)$)>>, <:expr<( $value$)>>
      | TInt32 -> var.t <- TInt32; <:expr<(new_int32_var $ExInt(_loc,string_of_int i)$)>>, <:expr<( $value$)>>
      | TInt64 -> var.t <- TInt64; <:expr<(new_int64_var $ExInt(_loc,string_of_int i)$)>>, <:expr<( $value$)>>
      | TFloat -> var.t <- TFloat64; <:expr<(new_double_var $ExInt(_loc,string_of_int i)$)>>,<:expr<( $value$)>>
      | TFloat64 -> var.t <- TFloat; <:expr<(new_double_var $ExInt(_loc,string_of_int i)$)>>,<:expr<( $value$)>>
      | _  -> (* assert false *) (); failwith "unknown var type"
    in
    incr arg_idx;
    Hashtbl.add !current_args (string_of_ident s) {n=i; var_type=var.t};
    arg_list := <:expr<spoc_set $ex1$ $ex2$>>:: !arg_list
    (*<:expr<new_var $ExInt(_loc,string_of_int i)$>>*)
  | _ -> failwith "error new_var"
*)

let remove_int_var var = 
  match var.e with 
  | Id (_loc, s)  -> 
    Hashtbl.remove !current_args (string_of_ident s);
  | CastId (_,Id(_loc, s) ) -> 
    Hashtbl.remove !current_args (string_of_ident s);
    (*<:expr<new_var $ExInt(_loc,string_of_int i)$>>*)
  | _ -> failwith "error new_var"

let rec parse_int i t= 
  match i.e with
  | Id (_loc,s)  -> 
    (
      let is_mutable = ref false in
     ( try 
	let var = 
	  Hashtbl.find !current_args (string_of_ident s)  in
	(
	  is_mutable := var.is_mutable;
	  match var.var_type with
	  | TUnknown -> var.var_type <- t
	  | x when x = t  -> ()
	  | _  -> (* assert false *) (); raise (TypeError (t, var.var_type, _loc)));
      with Not_found -> 
	(
	  try 
	    let c_const = Hashtbl.find !intrinsics_const (string_of_ident s) in
	    (match c_const.typ with
	    | x when x = t -> ()
	    | _  -> (* assert false *) (); raise (TypeError (t, c_const.typ, _loc)))
	  with Not_found ->
	    ((* assert false *) (); raise (Unbound_value ((string_of_ident s), _loc)))));
     (match i.t with
     | x when x = t -> () 

     | TUnknown  ->  i.t <- t; 
     | _ -> (* assert false *) (); raise (TypeError (t, i.t, i.loc)));
     if !is_mutable then
       <:expr< $(ExId (_loc, s))$.contents>>
     else
       <:expr< $(ExId (_loc, s))$>>)
  | Int (_loc, s) -> 
    ( match i.t with 
    | TInt  -> <:expr< $(ExInt (_loc, s))$>>
    | _ -> (* assert false *) (); raise (TypeError (t, i.t, i.loc)))
  | Int32 (_loc, s) -> 
    ( match i.t with 
    | TInt32  -> <:expr< $(ExInt32 (_loc, s))$>>
    | _ -> (* assert false *) (); raise (TypeError (t, i.t, i.loc)))
      
  | Int64 (_loc, s) -> 
    ( match i.t with 
    | TInt64  -> <:expr< $(ExInt64 (_loc, s))$>>
    | _ -> (* assert false *) (); raise (TypeError (t, i.t, i.loc)))
  | Plus (_loc, a, b) |Plus32 (_loc, a, b) | Plus64 (_loc, a, b)  -> 
    parse_body i
  | Min (_loc, a, b) |Min32 (_loc, a, b) | Min64 (_loc, a, b)  -> 
    parse_body i
  | Mul (_loc, a, b) |Mul32 (_loc, a, b) | Mul64 (_loc, a, b)  -> 
    parse_body i
  | Div (_loc, a, b) |Div32 (_loc, a, b) | Div64 (_loc, a, b)  -> 
    parse_body i
  | Mod (_loc, a, b) -> 
    parse_body i
  | Bind (_loc, var,y, z, is_mutable)  ->
    (
      let gen_z = parse_int z t in
      match var.e with
      | Id (_loc,s) ->  
	(<:expr<let $PaId(_loc,s)$ = $(parse_body y)$ in $gen_z$>>)
      | _ -> failwith "error parse_body Bind")
  | Ref (_loc, id) -> 
    <:expr< ! $parse_body id$>>
  | VecGet (_loc, vector, index)  -> 
    ( match i.t with
    | x when x = t -> ()
        (*| TUnknown -> i.t <- t*)
    | _  -> (* assert false *) (); raise (TypeError (t, i.t , _loc)));
    (*vector.t <- TVec i.t;*)
    (match vector.e with
    | Id (_loc,s)  -> 
      let var = (Hashtbl.find !current_args (string_of_ident s)) in
      let type_constraint = 
	let vec_typ_to_e k =
	   match k with 
	   | TInt32  -> <:ctyp<(int32, Bigarray.int32_elt) Spoc.Vector.vector>>
	   | TInt64  -> <:ctyp<(int64, Bigarray.int64_elt) Spoc.Vector.vector>>
	   | TInt ->  <:ctyp<(int, Bigarray.int_elt) Spoc.Vector.vector>>
	   |  _  ->  assert false 
		in
	(match var.var_type with
	| TVec k when k = t  -> (
	  vec_typ_to_e k
	)
	| TVec TUnknown -> var.var_type <- TVec t;
	  vec_typ_to_e t
	| _  -> (* assert false *) (); raise (TypeError (TVec t, var.var_type, _loc)))
	in
      <:expr<Spoc.Mem.get ($ExId(_loc,s)$:$type_constraint$) $parse_body index$>>
    | _  ->  (* assert false *) (); failwith "Unknwown vector");
  (*      <:expr<get_vec $parse_int vector$ $parse_body index$>>*)
  | _ -> (* assert false *) (); raise (TypeError (t, i.t, i.loc))


and parse_float f t = 
  match f.e with
  | App (_loc, e1, e2) ->
     parse_body f
  | CastId (tt,Id(_loc,s)) ->
    (if t = tt then
	(( let var = (
	     ( try Hashtbl.find !current_args (string_of_ident s) 
	       with _ -> (* assert false *) (); raise (Unbound_value ((string_of_ident s), _loc)))) in
	   match var.var_type with
	   | TUnknown ->
	     ( try 
		 (Hashtbl.find !current_args (string_of_ident s)).var_type <- t  
               with _ -> (* assert false *) (); raise (Unbound_value ((string_of_ident s),_loc)))
           | x when x = t  -> ()
           | _  -> (* assert false *) (); raise (TypeError (t, var.var_type, _loc)));
	 match f.t with
         | x when x = t ->  <:expr<$(ExId (_loc, s))$>>
         | TUnknown  ->  f.t <- t; 
           <:expr< $(ExId (_loc, s))$>>
         | _ -> (* assert false *) (); raise (TypeError (t, f.t, f.loc)))
     else 
	((* assert false *) (); raise (TypeError (t, f.t, f.loc))))
  | Id (_loc,s)  ->
    (let is_mutable = ref false in
     ( let var = (
	 ( try Hashtbl.find !current_args (string_of_ident s) 
	   with _ -> (* assert false *) (); raise (Unbound_value ((string_of_ident s),_loc)))) in
       is_mutable := var.is_mutable;
       match var.var_type with
       | TUnknown ->
	 ( try 
             (Hashtbl.find !current_args (string_of_ident s)).var_type <- t  
           with _ -> (* assert false *) (); raise (Unbound_value ((string_of_ident s),_loc)))
       | x when x = t  -> ()
       | _  -> (* assert false *) (); raise (TypeError (t, var.var_type, _loc)));
     (match f.t with
     | x when x = t -> ()
     | TUnknown  ->  f.t <- t
     | _ -> (* assert false *) (); raise (TypeError (t, f.t, f.loc))) ;
     if !is_mutable then
       <:expr< $(ExId (_loc, s))$.contents>>
     else
       <:expr< $(ExId (_loc, s))$>>)
 | Ref (_loc, id) -> 
    <:expr< ! $parse_body id$>>
  | Float (_loc, s)  -> 
    ( match f.t with 
    | TFloat  -> <:expr<$(ExFlo(_loc, s))$>>
    | _ -> (* assert false *) (); raise (TypeError (t, f.t, f.loc)))
  | Float32 (_loc, s) -> 
    ( match f.t with 
    | TFloat32  -> <:expr< $(ExFlo (_loc, s))$>>
    | _ -> (* assert false *) (); raise (TypeError (t, f.t, f.loc)))
  | Float64 (_loc, s) -> 
    ( match f.t with 
    | TFloat64  -> <:expr< $(ExFlo (_loc, s))$>>
    | _ -> (* assert false *) (); raise (TypeError (t, f.t, f.loc)))
  | PlusF (_loc, a, b) | PlusF32 (_loc, a, b) | PlusF64  (_loc, a, b)-> 
    parse_body f 
  | MinF (_loc, a, b) | MinF32 (_loc, a, b) | MinF64  (_loc, a, b)-> 
    parse_body f 
  | MulF (_loc, a, b) | MulF32 (_loc, a, b) | MulF64  (_loc, a, b)-> 
    parse_body f 
  | DivF (_loc, a, b) | DivF32 (_loc, a, b) | DivF64  (_loc, a, b)-> 
    parse_body f 
      
  | VecGet (_loc, vector, index)  -> 
    ( match f.t with
    | x when x = t -> ()
        (*| TUnknown -> i.t <- t*)
    | _  -> (* assert false *) (); raise (TypeError (t, f.t , _loc)));
    (*vector.t <- TVec i.t;*)
    (match vector.e with
    | Id (_loc,s)  -> 
      let var = (Hashtbl.find !current_args (string_of_ident s)) in
      let type_constraint = (match var.var_type with
	| TVec k when k = t  -> (
	  match k with 
  	  | TFloat -> <:ctyp<(float, Bigarray.float32_elt) Spoc.Vector.vector>>
	  | TFloat32 -> <:ctyp<(float, Bigarray.float32_elt) Spoc.Vector.vector>>
	  | TFloat64 -> <:ctyp<(float, Bigarray.float64_elt) Spoc.Vector.vector>>
	  |  _  ->  assert false 
	)
	| _  -> (* assert false *) (); raise (TypeError (TVec t, var.var_type, _loc))
      ) in
      <:expr<Spoc.Mem.get ($ExId(_loc,s)$:$type_constraint$) $parse_body index$>>
    | _  ->  (* assert false *) (); failwith "Unknwown vector");
  (*      <:expr<get_vec $parse_int vector$ $parse_body index$>>*)
  | ModuleAccess _ -> parse_body f
  | _ -> (*(* assert false *) (); *) 
    my_eprintf (Printf.sprintf "(* val %s *)\n%!" (k_expr_to_string f.e));
    raise (TypeError (t, f.t, f.loc))
    
and  parse_int2 i t= 
  match i.e with
  | Id (_loc,s) -> 
    (try 
       let var = (Hashtbl.find !current_args (string_of_ident s)) in
       (match var.var_type with
       (*| TUnknown  ->  var.var_type <- t*) 
       | x when x = t ->  ()
       | _  ->  (* assert false *) (); raise (TypeError (t, var.var_type, _loc)));
       if var.is_global then
	 <:expr<global_int_var $ExId(_loc,s)$>>
       else
	 <:expr<var  $ExInt(_loc, string_of_int var.n)$>>
     with
     | Not_found ->
       try 
	 let c_const = Hashtbl.find !intrinsics_const (string_of_ident s) in
	 match c_const.typ with
	 | x when x = t -> 
	   <:expr< intrinsics $ExStr(_loc, c_const.cuda_val)$ $ExStr(_loc, c_const.opencl_val)$>>
	 | _  -> (* assert false *) (); raise (TypeError (t, c_const.typ, _loc))
       with Not_found ->
	 ((* assert false *) (); raise (Unbound_value ((string_of_ident s),_loc))))
  | Ref (_, {loc=_; e=Id(_loc,s); t=_}) ->
    <:expr<global_int_var (fun () -> ! $ExId(_loc, s)$)>>
  | Int (_loc, s)  -> <:expr<spoc_int $(ExInt (_loc, s))$>>
  | Int32 (_loc, s)  -> <:expr<spoc_int32 $(ExInt32 (_loc, s))$>>
  | Int64 (_loc, s)  -> <:expr<spoc_int64 $(ExInt64 (_loc, s))$>>
    
  | Plus (_loc, a, b) | Plus32 (_loc, a, b)| Plus64 (_loc, a, b)  -> 
				parse_body2 i false
  | Min (_loc, a, b) | Min32 (_loc, a, b)| Min64 (_loc, a, b)  -> 
			      parse_body2 i false
  | Mul (_loc, a, b) | Mul32 (_loc, a, b)| Mul64 (_loc, a, b)  -> 
			      parse_body2 i false
  | Mod (_loc, a, b) -> 
    parse_body2 i false
  | Div (_loc, a, b) | Div32 (_loc, a, b)| Div64 (_loc, a, b)  -> 
				parse_body2 i false
  | Bind (_loc, var, y, z, is_mutable)  -> parse_body2 i false
  | VecGet (_loc, vector, index)  -> 
    ( match i.t with
    | x when x = t -> ()
    (*| TUnknown -> i.t <- t ;*)
    | _  -> (* assert false *) (); raise (TypeError (t, i.t , _loc)));
    (*vector.t <- TVec i.t;*)
    (match vector.e with
    | Id (_loc,s)  -> 
      let var = (Hashtbl.find !current_args (string_of_ident s)) in
      (match var.var_type with
      (*| TUnknown  ->  var.var_type <- TVec t*)
      | TVec x when x = t  -> ()
      | _  -> (* assert false *) (); raise (TypeError (TVec t, var.var_type, _loc))
      );
    | _  ->  failwith "Unknwown vector");
    <:expr<get_vec $parse_int2 vector (TVec t)$ $parse_int2 index TInt$>>
  | _ -> (* assert false *) (); raise (TypeError (t, i.t, i.loc))
    
and  parse_float2 f t= 
  match f.e with
  | App (_loc, e1, e2) ->
     parse_body2 f false
  | CastId(_, Id (_loc,s))  -> <:expr<double_var  ( $ExInt(_loc, string_of_int (Hashtbl.find !current_args (string_of_ident s)).n)$)>>
  | Id (_loc,s)  -> 
    (try 
       let var = (Hashtbl.find !current_args (string_of_ident s)) in
       (match var.var_type with
       (*| TUnknown  ->  var.var_type <- t*) 
       | x when x = t ->  ()
       | _  ->  (* assert false *) (); raise (TypeError (t, var.var_type, _loc)));
       if var.is_global then
	 <:expr<global_float_var $ExId(_loc,s)$>>
       else
	 <:expr<var  $ExInt(_loc, string_of_int var.n)$>>
     with
     | Not_found ->
       try 
	 let c_const = Hashtbl.find !intrinsics_const (string_of_ident s) in
	 match c_const.typ with
	 | x when x = t -> 
	   <:expr< intrinsics $ExStr(_loc, c_const.cuda_val)$ $ExStr(_loc, c_const.opencl_val)$>>
	 | _  -> (* assert false *) (); raise (TypeError (t, c_const.typ, _loc))
       with Not_found ->
	 ((* assert false *) (); raise (Unbound_value ((string_of_ident s),_loc))))
  | Ref (_, {loc=_; e=Id(_loc,s); t=_}) ->
    <:expr<global_float_var (fun () -> ! $ExId(_loc, s)$)>>
  | Float (_loc, s)  -> <:expr<spoc_float $(ExFlo(_loc, s))$>>
  | Float32 (_loc, s)  -> <:expr<spoc_float $(ExFlo(_loc, s))$>>
  | Float64 (_loc, s)  -> <:expr<spoc_double $(ExFlo(_loc, s))$>>
  | PlusF (_loc, a, b) |  PlusF32 (_loc, a, b) | PlusF64 (_loc, a, b)  -> 
    parse_body2 f false
  | MinF (_loc, a, b) |  MinF32 (_loc, a, b) | MinF64 (_loc, a, b)  -> 
    parse_body2 f false
	| MulF (_loc, a, b) |  MulF32 (_loc, a, b) | MulF64 (_loc, a, b)  -> 
    parse_body2 f false
  | DivF (_loc, a, b) |  DivF32 (_loc, a, b) | DivF64 (_loc, a, b)  -> 
    parse_body2 f false
  | VecGet (_loc, vector, index)  -> 
    ( match f.t with
    | x when x = t -> ()
        (*| TUnknown -> i.t <- t ;*)
    | _  -> (* assert false *) (); raise (TypeError (t, f.t , _loc)));
    (*vector.t <- TVec i.t;*)
    (match vector.e with
    | Id (_loc,s)  -> 
      let var = (Hashtbl.find !current_args (string_of_ident s)) in
      (match var.var_type with
      (*| TUnknown  ->  var.var_type <- TVec t*)
      | TVec x when x = t  -> ()
      | _  -> (* assert false *) (); raise (TypeError (TVec t, var.var_type, _loc))
      );
    | _  ->  failwith "Unknwown vector");
    <:expr<get_vec $parse_float2 vector (TVec t)$ $parse_int2 index TInt$>>
  | ModuleAccess _ -> parse_body2 f false
  | _  -> ( my_eprintf (Printf.sprintf "(* val %s *)\n%!" (k_expr_to_string f.e));
		raise (TypeError (t, f.t, f.loc));
		failwith "error parse_float2")	
    		
and parse_modacc m =
	let res = ref m in
  let rec aux m = 
			match m.e with
	| ModuleAccess (_loc, e1, e2) -> ExAcc(_loc, <:expr< $uid:e1$>>, aux e2)
	| Id _ -> parse_body m
	| App (_loc, e1, e2::[]) -> res:=e2; aux e1
	in
	(aux m), !res
		
and parse_app_ml a modu =
  match a.e with
  | Id (_loc,s) -> ExAcc(_loc, modu,ExId(_loc, s))
  | ModuleAccess (_loc, s, e) ->
    let e = parse_app_ml e (ExAcc(_loc, modu, ExId(_loc, IdUid(_loc,s)))) in e
	| App (_loc, e1, e2::[]) ->
		let rec aux a modu =
			match a.e with
		  | Id (_loc,s) -> ExApp(_loc, ExAcc(_loc, modu,ExId(_loc, s)), parse_body e2)
			| App (l, e1, e2::[]) -> ExApp (_loc, aux e1 modu, parse_body e2)
			| ModuleAccess (_loc, s, e) -> 
				ExAcc(_loc, modu, (parse_body a))
			| _ -> my_eprintf (Printf.sprintf "(* app %s *)\n%!" (k_expr_to_string a.e)); assert false 
		in
		ExApp(_loc, aux e1 modu, parse_body e2)
	| _ -> parse_body a 
  | _ -> 
    my_eprintf (Printf.sprintf"(* app %s *)\n%!" (k_expr_to_string a.e));
     assert false
		
				    
and parse_body body = 
  (*  let _loc = Loc.ghost in*)
  my_eprintf (Printf.sprintf"(* val %s *)\n%!" (k_expr_to_string body.e));
  match body.e with
  | Bind (_loc, var,y, z, is_mutable)  ->
    (
      let y = parse_body y in
      let gen_z = parse_body z in
       (*body.t <- z.t;*)
      match var.e with
      | Id (_loc,s) | CastId (_,Id(_loc,s)) ->
	if is_mutable then
      	(<:expr<let $PaId(_loc,s)$ = ref $y$ in $gen_z$>>)
	else
      	(<:expr<let $PaId(_loc,s)$ = $y$ in $gen_z$>>)
      | _ -> failwith "error parse_body Bind")
  | Plus (_loc, a,b) -> 
     (*body.t <- TInt32;*)
    ( <:expr<$(parse_int a TInt)$ + $(parse_int b TInt)$>>)
  | Plus32 (_loc, a,b) -> 
     (*body.t <- TInt32;*)
    ( <:expr<Int32.add $(parse_int a TInt32)$  $(parse_int b TInt32)$>>)
  | Plus64 (_loc, a, b) ->
    ( <:expr<Int64.add $(parse_int a TInt64)$  $(parse_int b TInt64)$>>)
  | PlusF (_loc, a,b) -> 
    ( <:expr<$(parse_float a TFloat32)$ +. $(parse_float b TFloat32)$>>)
  | PlusF32 (_loc, a,b) -> 
    	let a_ = (parse_float a TFloat32) in
		let b_ = (parse_float b TFloat32) in
		(if (a.t <> TFloat32 && a.t <> TUnknown) then
			raise (TypeError (TFloat32, a.t, a.loc)));
		(if (b.t <> TFloat32 && b.t <> TUnknown) then
			raise (TypeError (TFloat32, b.t, b.loc)));
    ( <:expr<$a_$ +. $b_$>>)
  | PlusF64 (_loc, a,b) -> 
    	let a_ = (parse_float a TFloat64) in
		let b_ = (parse_float b TFloat64) in
		(if (a.t <> TFloat64 && a.t <> TUnknown) then
			raise (TypeError (TFloat64, a.t, a.loc)));
		(if (b.t <> TFloat64 && b.t <> TUnknown) then
			raise (TypeError (TFloat64, b.t, b.loc)));
    ( <:expr<$a_$ +. $b_$>>)

  | Min (_loc, a,b) -> 
     (*body.t <- TInt32;*)
    ( <:expr<$(parse_int a TInt)$ - $(parse_int b TInt)$>>)
  | Min32 (_loc, a,b) -> 
     (*body.t <- TInt32;*)
    ( <:expr<Int32.min $(parse_int a TInt32)$  $(parse_int b TInt32)$>>)
  | Min64 (_loc, a, b) ->
    ( <:expr<Int64.min $(parse_int a TInt64)$  $(parse_int b TInt64)$>>)
  | MinF (_loc, a,b) -> 
    ( <:expr<$(parse_float a TFloat)$ -. $(parse_float b TFloat)$>>)
  | MinF32 (_loc, a,b) -> 
     	let a_ = (parse_float a TFloat32) in
		let b_ = (parse_float b TFloat32) in
		(if (a.t <> TFloat32 && a.t <> TUnknown) then
			raise (TypeError (TFloat32, a.t, a.loc)));
		(if (b.t <> TFloat32 && b.t <> TUnknown) then
			raise (TypeError (TFloat32, b.t, b.loc)));
    ( <:expr<$a_$ -. $b_$>>)
  | MinF64 (_loc, a,b) -> 
    	let a_ = (parse_float a TFloat64) in
		let b_ = (parse_float b TFloat64) in
		(if (a.t <> TFloat64 && a.t <> TUnknown) then
			raise (TypeError (TFloat64, a.t, a.loc)));
		(if (b.t <> TFloat64 && b.t <> TUnknown) then
			raise (TypeError (TFloat64, b.t, b.loc)));
    ( <:expr<$a_$ -. $b_$>>)


  | Mul (_loc, a,b) -> 
     (*body.t <- TInt32;*)
    ( <:expr<$(parse_int a TInt)$ * $(parse_int b TInt)$>>)
  | Mul32 (_loc, a,b) -> 
     (*body.t <- TInt32;*)
    ( <:expr<Int32.mul $(parse_int a TInt32)$  $(parse_int b TInt32)$>>)
  | Mul64 (_loc, a, b) ->
    ( <:expr<Int64.mul $(parse_int a TInt64)$  $(parse_int b TInt64)$>>)
  | MulF (_loc, a,b) -> 
    ( <:expr<$(parse_float a TFloat32)$ *. $(parse_float b TFloat32)$>>)
  | MulF32 (_loc, a,b) -> 
		 	let a_ = (parse_float a TFloat32) in
		let b_ = (parse_float b TFloat32) in
		(if (a.t <> TFloat32 && a.t <> TUnknown) then
			raise (TypeError (TFloat32, a.t, a.loc)));
		(if (b.t <> TFloat32 && b.t <> TUnknown) then
			raise (TypeError (TFloat32, b.t, b.loc)));
    ( <:expr<$a_$ *. $b_$>>)
  | MulF64 (_loc, a,b) -> 
    	let a_ = (parse_float a TFloat64) in
		let b_ = (parse_float b TFloat64) in
		(if (a.t <> TFloat64 && a.t <> TUnknown) then
			raise (TypeError (TFloat64, a.t, a.loc)));
		(if (b.t <> TFloat64 && b.t <> TUnknown) then
			raise (TypeError (TFloat64, b.t, b.loc)));
    ( <:expr<$a_$ *. $b_$>>)

  | Div (_loc, a,b) -> 
     (*body.t <- TInt32;*)
    ( <:expr<$(parse_int a TInt)$ / $(parse_int b TInt)$>>)
  | Div32 (_loc, a,b) -> 
     (*body.t <- TInt32;*)
    ( <:expr<Int32.div $(parse_int a TInt32)$  $(parse_int b TInt32)$>>)
  | Div64 (_loc, a, b) ->
    ( <:expr<Int64.div $(parse_int a TInt64)$  $(parse_int b TInt64)$>>)
  | DivF (_loc, a,b) -> 
    ( <:expr<$(parse_float a TFloat32)$ /. $(parse_float b TFloat32)$>>)
  | DivF32 (_loc, a,b) -> 
    	let a_ = (parse_float a TFloat32) in
		let b_ = (parse_float b TFloat32) in
		(if (a.t <> TFloat32 && a.t <> TUnknown) then
			raise (TypeError (TFloat32, a.t, a.loc)));
		(if (b.t <> TFloat32 && b.t <> TUnknown) then
			raise (TypeError (TFloat32, b.t, b.loc)));
    ( <:expr<$a_$ /. $b_$>>)
  | DivF64 (_loc, a,b) -> 
   	let a_ = (parse_float a TFloat64) in
		let b_ = (parse_float b TFloat64) in
		(if (a.t <> TFloat64 && a.t <> TUnknown) then
			raise (TypeError (TFloat64, a.t, a.loc)));
		(if (b.t <> TFloat64 && b.t <> TUnknown) then
			raise (TypeError (TFloat64, b.t, b.loc)));
    ( <:expr<$a_$ /. $b_$>>)


  | Mod (_loc, a,b) -> 
     (*body.t <- TInt32;*)
    ( <:expr<$(parse_int a TInt)$ mod $(parse_int b TInt)$>>)


  | Id (_loc,s) -> 
    begin
      try
	let var = (Hashtbl.find !current_args (string_of_ident s)) in
	if var.is_mutable then
	  <:expr< $ExId(_loc,s)$.contents>>
	else
	  <:expr<$ExId(_loc,s)$>>
      with 
	Not_found ->
	  <:expr<$ExId(_loc,s)$>>
    end
  | CastId (_,Id(_loc,s)) -> 
    <:expr<$ExId(_loc,s)$>>
  | Int (_loc, i)  -> <:expr<$ExInt(_loc, i)$>>
  | Int32 (_loc, i)  -> <:expr<$ExInt32(_loc, i)$>>
  | Int64 (_loc, i)  -> <:expr<$ExInt64(_loc, i)$>>
  | Float (_loc, f)
  | Float32 (_loc,f)
  | Float64 (_loc,f)  -> <:expr<$ExFlo(_loc, f)$>>
  | Seq (_loc, x, y) -> 
    let x = parse_body x in
    let y = parse_body y in
    <:expr<$x$; $y$>>
  | End(_loc, x)  -> 
    let res = parse_body x in
        (*body.t <- x.t;*) 
 (*   return_type := x.t;*) <:expr<$res$>> 
  | VecSet (_loc, vector, value)  -> 								
    let gen_value = parse_body value in
    
          (*body.t <- TUnit;
            vector.t <- value.t;*)
    (match vector.e with
    | VecGet (_, v,idx)  -> 
      (*v.t <- TVec value.t;*)
      let vec_typ_to_e t = 
	(match t with 
	| TInt -> <:ctyp<(int, Bigarray.int_elt) Spoc.Vector.vector>>
        | TInt32  -> <:ctyp<(int32, Bigarray.int32_elt) Spoc.Vector.vector>>
        | TInt64  -> <:ctyp<(int64, Bigarray.int64_elt) Spoc.Vector.vector>>
        | TFloat -> <:ctyp<(float, Bigarray.float32_elt) Spoc.Vector.vector>>
	| TFloat32 -> <:ctyp<(float, Bigarray.float32_elt) Spoc.Vector.vector>>
        | TFloat64 -> <:ctyp<(float, Bigarray.float64_elt) Spoc.Vector.vector>>
	|  _  ->  assert false 
        ) in
      (match v.e with
      | Id (_loc,s)  -> 
	let var = (Hashtbl.find !current_args (string_of_ident s)) in
	let type_constaint = 
	  (match var.var_type with
	  (*| TUnknown  ->  var.var_type <- TVec value.t*)
	  | TVec k when k = value.t  -> 
	    vec_typ_to_e k
	  | _ ->   
	    (match value.t, var.var_type with
	    | TUnknown, TUnknown ->
	      ((* assert false *) (); raise (TypeError (TVec value.t, var.var_type, _loc)))
	    | TUnknown, TVec tt ->
	      (var.var_type <- TVec tt;
	       vec_typ_to_e tt)
	    | tt, TVec TUnknown ->
	      (var.var_type <- TVec tt;
	       vec_typ_to_e tt)
	    | _ ->
	      ((* assert false *) (); raise (TypeError (TVec value.t, var.var_type, _loc))))
	  )
	in
	
	<:expr<Spoc.Mem.set ($parse_body v$:$type_constaint$)
	  $parse_body idx$ $gen_value$>>
      | _  ->  failwith "Unknwown vector");
    | _  -> failwith (Printf.sprintf "erf %s" (k_expr_to_string vector.e)) ); 
    
  | VecGet(_loc, vector, index)  -> 
    ignore(parse_body vector);
           (*body.t <- (match vector.t with 
             | TVec k -> k
             | _ -> raise (TypeError (TVec TUnknown, vector.t, _loc)));*)
    (match vector.e with
    | Id (_loc, s)-> 
      let var = (Hashtbl.find !current_args (string_of_ident s)) in
      let type_constraint = 
        (match var.var_type with
                    (*| TUnknown  ->  var.var_type <- TVec value.t*)
        | TVec k -> 
          (match k with 
	  | TInt -> <:ctyp<(int, Bigarray.int_elt) Spoc.Vector.vector>>
          | TInt32  -> <:ctyp<(int32, Bigarray.int32_elt) Spoc.Vector.vector>>
          | TInt64  -> <:ctyp<(int64, Bigarray.int64_elt) Spoc.Vector.vector>>
          | TFloat -> <:ctyp<(float, Bigarray.float32_elt) Spoc.Vector.vector>>
	  | TFloat32 -> <:ctyp<(float, Bigarray.float32_elt) Spoc.Vector.vector>>
          | TFloat64 -> <:ctyp<(float, Bigarray.float64_elt) Spoc.Vector.vector>>
	  | TBool | TVec _ ->  assert false 
	  | TUnknown ->  assert false 
	  |	_  ->  assert false 
          )
        | _  -> (* assert false *) (); failwith "strange vector"
        )in
      <:expr<Spoc.Mem.get ($ExId(_loc,s)$:$type_constraint$) $parse_body index$>>
    | _  -> (* assert false *) (); failwith "strange vector")
  | BoolEq (_loc, a, b) ->
    (<:expr<$parse_int a TInt$ = $parse_int b TInt$>>)
  | BoolEq32 (_loc, a, b) ->
    (<:expr<$parse_int a TInt32$ = $parse_int b TInt32$>>)
  | BoolEq64 (_loc, a, b) ->
    (<:expr<$parse_int a TInt64$ = $parse_int b TInt64$>>)
  | BoolEqF (_loc, a, b) ->
    (<:expr<$parse_float a TFloat32$ = $parse_float b TFloat32$>>)
  | BoolEqF64 (_loc, a, b) ->
    (<:expr<$parse_float a TFloat64$ = $parse_float b TFloat64$>>)

  | BoolLt (_loc, a, b) ->
    (<:expr<$parse_int a TInt$ < $parse_int b TInt$>>)
  | BoolLt32 (_loc, a, b) ->
    (<:expr<$parse_int a TInt32$ < $parse_int b TInt32$>>)
  | BoolLt64 (_loc, a, b) ->
    (<:expr<$parse_int a TInt64$ < $parse_int b TInt64$>>)
  | BoolLtF (_loc, a, b) ->
    (<:expr<$parse_float a TFloat32$ < $parse_float b TFloat32$>>)
  | BoolLtF64 (_loc, a, b) ->
    (<:expr<$parse_float a TFloat64$ < $parse_float b TFloat64$>>)

  | BoolGt (_loc, a, b) ->
    (<:expr<$parse_int a TInt$ > $parse_int b TInt$>>)
  | BoolGt32 (_loc, a, b) ->
    (<:expr<$parse_int a TInt32$ > $parse_int b TInt32$>>)
  | BoolGt64 (_loc, a, b) ->
    (<:expr<$parse_int a TInt64$ > $parse_int b TInt64$>>)
  | BoolGtF (_loc, a, b) ->
    (<:expr<$parse_float a TFloat32$ > $parse_float b TFloat32$>>)
  | BoolGtF64 (_loc, a, b) ->
    (<:expr<$parse_float a TFloat64$ > $parse_float b TFloat64$>>)

  | BoolLtE (_loc, a, b) ->
    (<:expr<$parse_int a TInt$ <= $parse_int b TInt$>>)
  | BoolLtE32 (_loc, a, b) ->
    (<:expr<$parse_int a TInt32$ <= $parse_int b TInt32$>>)
  | BoolLtE64 (_loc, a, b) ->
    (<:expr<$parse_int a TInt64$ <= $parse_int b TInt64$>>)
  | BoolLtEF (_loc, a, b) ->
    (<:expr<$parse_float a TFloat32$ <= $parse_float b TFloat32$>>)
  | BoolLtEF64 (_loc, a, b) ->
    (<:expr<$parse_float a TFloat64$ <= $parse_float b TFloat64$>>)

  | BoolGtE (_loc, a, b) ->
    (<:expr<$parse_int a TInt$ >= $parse_int b TInt$>>)
  | BoolGtE32 (_loc, a, b) ->
    (<:expr<$parse_int a TInt32$ >= $parse_int b TInt32$>>)
  | BoolGtE64 (_loc, a, b) ->
    (<:expr<$parse_int a TInt64$ >= $parse_int b TInt64$>>)
  | BoolGtEF (_loc, a, b) ->
    (<:expr<$parse_float a TFloat32$ >= $parse_float b TFloat32$>>)
  | BoolGtEF64 (_loc, a, b) ->
    (<:expr<$parse_float a TFloat64$ >= $parse_float b TFloat64$>>)

  | BoolOr (_loc, a, b) ->
    (<:expr<$parse_body a$ || $parse_body b$>>)
  | BoolAnd (_loc, a, b) ->
    (<:expr<$parse_body a$ &&  $parse_body b$>>)
    
 | Ife (_loc, cond, cons1, cons2) ->
    let cond  = parse_body cond in 
    let cons1 = parse_body cons1 in 
    let cons2 = parse_body cons2 in 
    (<:expr<if $cond$ then $cons1$ else $cons2$>>)
 | If (_loc, cond, cons1) ->
    let cond  = parse_body cond in 
    let cons1 = parse_body cons1 in 
    (<:expr<if $cond$ then $cons1$>>)
 | DoLoop (_loc, ({t=_;e=Id(_,s) ;loc=_} as id), min, max, body) ->
   let var = (Hashtbl.find !current_args (string_of_ident s)) in
   var.var_type <- TInt;
   let min = (parse_body min) in
   let max = (parse_body max) in
   let body = parse_body body in
   let string_of_id id =
     match id with
     | Id (_,s ) -> string_of_ident s
     | _ ->  assert false 
   in
   (<:expr<for $(string_of_id id.e)$ = $min$ to $max$ do $body$ done>>)
 | While (_loc, cond, body) ->
   let cond = parse_body cond in
   let body = parse_body body in
   (<:expr< while $cond$ do $body$ done>>)
 | Seqs exprs ->
    let _loc = body.loc in
    List.fold_left     
      (fun s e ->
	let e2 = parse_body e in
	<:expr< $e2$; $s$>>) 
    <:expr< >>     
      exprs
  | App (_loc, e1, e2) ->
		let rec aux e2 = 
			match e2 with
			| t::[] -> parse_body t
			| t::q -> ExApp(_loc, (aux q), (parse_body t))
		in 
		let e2 = aux e2 in
(*		let e2 = List.fold_left (fun expr elt -> <:expr< $expr$ $parse_body elt$>>) (parse_body (List.hd e2)) (List.tl e2) in *)
		let e1 = parse_body e1 in	
    <:expr<$e1$ $e2$>>
  | Open (_loc, id, e) ->
		let rec aux = function
		| IdAcc (l,a,b) -> aux a; aux b
		| IdUid (l,s) -> open_module s l
		| _ -> (* assert false *) ()
		in
		aux id;
    let e = <:expr<let open $id$ in $parse_body e$>> in
		let rec aux = function
		| IdAcc (l,a,b) -> aux a; aux b
		| IdUid (l,s) -> close_module s
		| _ -> (* assert false *) ()
		in
		aux id;
		e
  | Noop -> 
    let _loc = body.loc in
    <:expr< ()>>
  | Acc (_loc, var, value) ->
    (match var.t with
    | TVec _ ->
      parse_body {
	t = body.t; 
	e = (VecSet (_loc, var, value));
	loc = body.loc} 
    | _ -> 
      <:expr<$parse_body var$ <- $parse_body value$>>
    )
  | Ref (_, {loc=_; e=Id(_loc,s); t=_}) ->
    <:expr< ! $ExId(_loc, s)$>>
  | ModuleAccess (_loc, s, e) -> 
    let e = parse_app_ml e <:expr< $uid:s$>> in
  	e
  | _ -> (* assert false *) (); failwith "unimplemented yet"

and open_module m_ident  _loc = 
  my_eprintf (Printf.sprintf "opening module %s\n%!" m_ident);
	
	(match m_ident with
	| "Float64" -> 
		if not (List.mem ex64 !extensions)  then 
			(
				extensions := ex64:: !extensions;
				Printf.eprintf "%s\n%!" ("\027[32m Warning \027[00m : kernel uses"^
					"\027[34m double precision floating point \027[00m"^
					"extension, make sure your device is compatible")
			)
	| _ -> ());
		let m =
    try Hashtbl.find modules (m_ident)
    with
    | _ -> (* assert false *) (); raise (Unbound_module (m_ident, _loc))
  in
  List.iter (fun (typ,s,nb_args,cuda_s, opencl_s) ->
    (Hashtbl.add !intrinsics_fun (s) {nb_args=2; cuda_val = cuda_s; opencl_val = opencl_s; typ=typ})) m.mod_functions;
  List.iter (fun (typ,s,cuda_s, opencl_s) ->
    (Hashtbl.add !intrinsics_const (s) {nb_args=0; cuda_val = cuda_s; opencl_val = opencl_s; typ=typ})) m.mod_constants ;
  Hashtbl.iter (fun name intern_m-> Hashtbl.add modules name intern_m) m.mod_modules 


and close_module m_ident = 
  my_eprintf (Printf.sprintf "closing module %s\n%!" m_ident);
  try
		let m =
     Hashtbl.find modules (m_ident)    
  in
  List.iter (fun (typ,s,nb_args,cuda_s, opencl_s) ->
    (Hashtbl.remove !intrinsics_fun (s))) m.mod_functions;
  List.iter (fun (typ,s,cuda_s, opencl_s) ->
    (Hashtbl.remove !intrinsics_const (s))) m.mod_constants;
  Hashtbl.iter (fun name intern_m-> Hashtbl.remove modules name) m.mod_modules 
	with
    | _ -> () (*(* assert false *) (); raise (Unbound_module (m_ident))*)
    

and typer_app e1 (e2 : kexpr list) t =
  let  typ, loc  = 
    let rec aux e1 =
      match e1.e with
      | Id (_l, s) -> (try (Hashtbl.find !intrinsics_fun (string_of_ident s)).typ , _l
			with |_ -> 
				typer e1 t; e1.t, _l); (*my_eprintf "%s\n" (string_of_ident s); (* assert false *) ()) *)
      | CastId (typ, Id (_l, s)) -> 
				(	let var  = Hashtbl.find !intrinsics_fun (string_of_ident s)
				in
				var.typ , _l 
			)
      | ModuleAccess (_l, s, e) ->
				open_module s _l;
				let typ, loc = aux e in
				close_module s;
				typ, loc
			| _ ->  typer e1 t; e1.t, e1.loc
      | _ ->  assert false (*typer e1 t; e1.t, Loc.ghost*)
   in 
		aux e1
  in
  let ret = ref TUnit in
	let rec aux2 typ expr =
		match typ, expr with
		| TApp (t1, t2), e::[] -> typer e t1; ret := t2
		| _ , [] -> (* assert false *) ()
	in		
  let rec aux typ1 e =
		match typ1,e with
    | (TApp (t1, (TApp (_,_) as t2)), 
			App ( l , e1, (t::(tt::qq as q) as e2))) ->
			aux2 t2 e2;
			typer e1 t1;
			if e1.t <> t1  then ( (* assert false *) (); raise (TypeError (t1, e1.t, l)) );
      e1.t <- t1;
	  | ((TApp (TApp(t1, t3) as t, t2)), 
       (App (l, e1, e2::[]))) ->
      assert (t3 = t2);
			ret := t2;
			typer e2 t1;			
			if e2.t <> t1 && e2.t <> TUnknown then ( (* assert false *) (); raise (TypeError (t1, e2.t, l)) );
      e2.t <- t1;
      e1.t <- t;
		| (TApp(t1, t2), App(_, _, e2::[] ) )->   
			my_eprintf (Printf.sprintf"(* typ %s +++-> %s *)\n%!" (k_expr_to_string e2.e) (ktyp_to_string e2.t)) ; 
			ret := e2.t;
			typer e2 t1;
  	| (t1 , App(_, _, e2::[] ) )->   
			my_eprintf (Printf.sprintf"(* typ %s +++-> %s *)\n%!" (k_expr_to_string e2.e) (ktyp_to_string e2.t)) ; 
			ret := t1;
			typer e2 t1;
    | _ -> (* assert false *) ();
  in aux typ (App (loc,e1,e2));
  let rec aux typ =
		match typ with
		| TApp(t1,t2) -> t2
		| t -> t
	in 
	my_eprintf (Printf.sprintf"(* >>>>>>>>>>>> typ %s *)\n%!" (ktyp_to_string typ)) ; 
	let t = aux typ  in
	my_eprintf (Printf.sprintf"(* >>>>>>>>>>>> typ %s *)\n%!" (ktyp_to_string t)) ;
	t
  
and typer body t = 
  (*  let _loc = Loc.ghost in*)
  my_eprintf (Printf.sprintf"(* typ %s -> %s  ??? %s *)\n%!" (k_expr_to_string body.e) (ktyp_to_string t)  (ktyp_to_string body.t)); 
  match body.e with
  | Bind (_loc, var,y, z, is_mutable)  ->
    (
      typer y TUnknown;
      (match var.e with
      | Id (_loc,s)  ->
        (incr arg_idx;
         Hashtbl.add !current_args (string_of_ident s) 
	   {n = !arg_idx; var_type = y.t;
	    is_mutable = is_mutable;
	    read_only = false;
	    write_only = false;
	    is_global = false;}
	)
      | CastId (tt, Id(_loc,s)) ->
	let ty = match y.t with 
	  | TFloat -> TFloat64
	  | _ -> y.t in
        (incr arg_idx;
         Hashtbl.add !current_args (string_of_ident s) 
	   {n = !arg_idx; var_type = ty;
	    is_mutable = is_mutable;
	    read_only = false;
	    write_only = false;
	    is_global=false;})
      | _ -> ((* assert false *) ())
      );
      var.t <- y.t;
      body.t <- y.t;
      typer z t;
    )   
  | Plus (_loc, a,b) -> 
    (match t with
    | TUnknown | TInt -> ()
    | _ -> (* assert false *) ());
    (typer a TInt; 
     typer b TInt;
     body.t <- TInt;)
  | Plus32 (_loc, a,b) -> 
    (match t with
    | TUnknown | TInt32 -> ()
    | _ -> (* assert false *) ());
    (typer a TInt32; 
     typer b TInt32;
     body.t <- TInt32;)
  | Plus64 (_loc, a,b) -> 
    (match t with
    | TUnknown | TInt64 -> ()
    | _ -> (* assert false *) ());    (typer a TInt64; 
     typer b TInt64;
     body.t <- TInt64;)
  | PlusF (_loc, a,b) -> 
    (match t with
    | TUnknown | TFloat32 -> ()
    | _ -> (* assert false *) ());    
    (typer a TFloat32; 
     typer b TFloat32;
     body.t <- TFloat32;)
  | PlusF32 (_loc, a,b) -> 
    (match t with
    | TUnknown | TFloat32 -> ()
    | _ -> (* assert false *) ());
    (typer a TFloat32; 
     typer b TFloat32;
     body.t <- TFloat32;)
  | PlusF64 (_loc, a,b) ->
    (match t with
    | TUnknown | TFloat64 -> ()
    | _ -> (* assert false *) ()); 
    (typer a TFloat64; 
     typer b TFloat64;
     body.t <- TFloat64;) 

  | Min (_loc, a,b) -> 
    (match t with
    | TUnknown | TInt -> ()
    | _ -> (* assert false *) ());
    (typer a TInt; 
     typer b TInt;
     body.t <- TInt;)
  | Min32 (_loc, a,b) -> 
    (match t with
    | TUnknown | TInt32 -> ()
    | _ -> (* assert false *) ());
    (typer a TInt32; 
     typer b TInt32;
     body.t <- TInt32;)
  | Min64 (_loc, a,b) -> 
    (match t with
    | TUnknown | TInt64 -> ()
    | _ -> (* assert false *) ());    (typer a TInt64; 
     typer b TInt64;
     body.t <- TInt64;)
  | MinF (_loc, a,b) -> 
    (match t with
    | TUnknown | TFloat32 -> ()
    | _ -> (* assert false *) ());    
    (typer a TFloat32; 
     typer b TFloat32;
     body.t <- TFloat32;)
  | MinF32 (_loc, a,b) -> 
    (match t with
    | TUnknown | TFloat32 -> ()
    | _ -> (* assert false *) ());
    (typer a TFloat32; 
     typer b TFloat32;
     body.t <- TFloat32;)
  | MinF64 (_loc, a,b) ->
    (match t with
    | TUnknown | TFloat64 -> ()
    | _ -> (* assert false *) ()); 
    (typer a TFloat64; 
     typer b TFloat64;
     body.t <- TFloat64;) 

  | Mul (_loc, a,b) -> 
    (match t with
    | TUnknown | TInt -> ()
    | _ -> (* assert false *) ());
    (typer a TInt; 
     typer b TInt;
     body.t <- TInt;)
  | Mul32 (_loc, a,b) -> 
    (match t with
    | TUnknown | TInt32 -> ()
    | _ -> (* assert false *) ());
    (typer a TInt32; 
     typer b TInt32;
     body.t <- TInt32;)
  | Mul64 (_loc, a,b) -> 
    (match t with
    | TUnknown | TInt64 -> ()
    | _ -> (* assert false *) ());    (typer a TInt64; 
     typer b TInt64;
     body.t <- TInt64;)
  | MulF (_loc, a,b) -> 
    (match t with
    | TUnknown | TFloat32 -> ()
    | _ -> (* assert false *) ());    
    (typer a TFloat32; 
     typer b TFloat32;
     body.t <- TFloat32;)
  | MulF32 (_loc, a,b) -> 
    (match t with
    | TUnknown | TFloat32 -> ()
    | _ -> (* assert false *) ());
    (typer a TFloat32; 
     typer b TFloat32;
     body.t <- TFloat32;)
  | MulF64 (_loc, a,b) ->
    (match t with
    | TUnknown | TFloat64 -> ()
    | _ -> (* assert false *) ()); 
    (typer a TFloat64; 
     typer b TFloat64;
     body.t <- TFloat64;) 

  | Div (_loc, a,b) -> 
    (match t with
    | TUnknown | TInt -> ()
    | _ -> (* assert false *) ());
    (typer a TInt; 
     typer b TInt;
     body.t <- TInt;)
  | Div32 (_loc, a,b) -> 
    (match t with
    | TUnknown | TInt32 -> ()
    | _ -> (* assert false *) ());
    (typer a TInt32; 
     typer b TInt32;
     body.t <- TInt32;)
  | Div64 (_loc, a,b) -> 
    (match t with
    | TUnknown | TInt64 -> ()
    | _ -> (* assert false *) ());    (typer a TInt64; 
     typer b TInt64;
     body.t <- TInt64;)
  | DivF (_loc, a,b) -> 
    (match t with
    | TUnknown | TFloat32 -> ()
    | _ -> (* assert false *) ());    
    (typer a TFloat32; 
     typer b TFloat32;
     body.t <- TFloat32;)
  | DivF32 (_loc, a,b) -> 
    (match t with
    | TUnknown | TFloat32 -> ()
    | _ -> (* assert false *) ());
    (typer a TFloat32; 
     typer b TFloat32;
     body.t <- TFloat32;)
  | DivF64 (_loc, a,b) ->
    (match t with
    | TUnknown | TFloat64 -> ()
    | _ -> (* assert false *) ()); 
    (typer a TFloat64; 
     typer b TFloat64;
     body.t <- TFloat64;) 
  | Mod (_loc, a,b) -> 
    (match t with
    | TUnknown | TInt -> ()
    | _ -> (* assert false *) ());
    (typer a TInt; 
     typer b TInt;
     body.t <- TInt;)
  | Id (_loc,s) ->  ( 
    try 
      let var = Hashtbl.find !current_args (string_of_ident s) in
      if t != TUnknown && t <> TVec TUnknown then
	if var.var_type = TUnknown then
	  (var.var_type <- t;
	   body.t <- t)
	else
	  if t <> var.var_type then
	    ( (* assert false *) (); raise (TypeError (t, var.var_type, _loc)))
	  else 
	    body.t <- t
    with Not_found ->
      try 
	let c_const = Hashtbl.find !intrinsics_const (string_of_ident s) in
	if t = TUnknown then
	  (body.t <- c_const.typ;)
	else if t <> c_const.typ then
	  ((* assert false *) (); raise (TypeError (c_const.typ, t, _loc)))
      with _ -> 
	try ignore(Hashtbl.find !intrinsics_fun (string_of_ident s) )
	with _ ->
          Hashtbl.add !current_args (string_of_ident s) 
	    {n = -1; var_type = t;
	     is_mutable = false;
	     read_only = true;
	     write_only = false;
	     is_global = true;}	    
	    (*	  (* assert false *) (); raise (Unbound_value (string_of_ident s)) *)
  )
  | CastId (tt, Id(_loc,s)) -> 
    ( let var = 
	( try Hashtbl.find !current_args (string_of_ident s) 
	  with _  -> (* assert false *) (); raise (Unbound_value ((string_of_ident s),_loc))) in
      var.var_type <- t;
      body.t <- t)
  | Int (_loc, i)  -> body.t <- TInt 
  | Int32 (_loc, i)  -> body.t <- TInt32 
  | Int64 (_loc, i)  -> body.t <- TInt64 
  | Float (_loc, f)  -> body.t <- TFloat32
  | Float32 (_loc, f) -> body.t <- TFloat32
  | Float64 (_loc, f) -> body.t <- TFloat64
  | Seq (_loc, x, y) -> 
    (typer x TUnit;
     typer y t; 
     body.t <- y.t)
 | Seqs exprs ->
    let rec aux = function
      | [] -> ()
      | e::[] ->
	typer e TUnknown
      | e::q -> typer e TUnit; aux q
    in 
    aux exprs
 | End(_loc, x)  -> ()

 | Acc (_loc, var, value) ->
   typer var TUnknown;
   (match var.t with
   | TVec _ -> 
     typer {
     t = body.t; 
     e = (VecSet (_loc, var, value));
     loc = body.loc} 
     t;
     body.t <- TUnit
   | _ -> 
     typer var TUnknown; 
     typer value var.t;
     body.t <- TUnit
   )

 | VecSet (_loc, vector, value)  -> 
    ((match value.e with
    | Id (_loc, s) ->
      let var = 
        ( try Hashtbl.find !current_args (string_of_ident s)
          with _ -> (* assert false *) (); raise (Unbound_value ((string_of_ident s),_loc))) in
      typer value var.var_type;
      typer vector (value.t)
    | _ -> typer value t);
     typer vector ( value.t);
(*     typer vector t; *)
     (match vector.e with
     | VecGet(_,v,_) -> 
       (match v.e with 
       | Id (_loc,s)  -> 
         ( let var = 
             ( try Hashtbl.find !current_args (string_of_ident s)
               with _ -> (* assert false *) (); raise (Unbound_value ((string_of_ident s), _loc))) in
           var.var_type <- TVec value.t)
       | _  -> () )
     | _  -> () );
     vector.t <- value.t;
     body.t <- TUnit;)        
  | VecGet(_loc, vector, index)  -> 
    (typer vector (TVec t);
     typer index TInt;
     (match vector.e with
     | Id (_loc,s)  -> 
       ( let var = 
           ( try Hashtbl.find !current_args (string_of_ident s)
             with 
             | _ -> (* assert false *) (); raise (Unbound_value ((string_of_ident s), _loc))) in
         var.var_type <- TVec t)
     | _  ->  () );
     vector.t <- TVec t;
     body.t <- t) 
 | BoolEq(_loc, a, b) ->
   typer a TInt;
   typer b TInt;
   body.t <- TBool;
 | BoolEq32(_loc, a, b) ->
   typer a TInt32;
   typer b TInt32;
   body.t <- TBool;
 | BoolEq64(_loc, a, b) ->
   typer a TInt64;
   typer b TInt64;
   body.t <- TBool;
 | BoolEqF(_loc, a, b) ->
   typer a TFloat32;
   typer b TFloat32;
   body.t <- TBool;
 | BoolEqF64(_loc, a, b) ->
   typer a TFloat64;
   typer b TFloat64;
   body.t <- TBool;
 | BoolLt(_loc, a, b) ->
   typer a TInt;
   typer b TInt;
   body.t <- TBool;
 | BoolLt32(_loc, a, b) ->
   typer a TInt32;
   typer b TInt32;
   body.t <- TBool;
 | BoolLt64(_loc, a, b) ->
   typer a TInt64;
   typer b TInt64;
   body.t <- TBool;
 | BoolLtF(_loc, a, b) ->
   typer a TFloat32;
   typer b TFloat32;
   body.t <- TBool;
 | BoolLtF64(_loc, a, b) ->
   typer a TFloat64;
   typer b TFloat64;
   body.t <- TBool;  

 | BoolGt(_loc, a, b) ->
   typer a TInt;
   typer b TInt;
   body.t <- TBool;
 | BoolGt32(_loc, a, b) ->
   typer a TInt32;
   typer b TInt32;
   body.t <- TBool;
 | BoolGt64(_loc, a, b) ->
   typer a TInt64;
   typer b TInt64;
   body.t <- TBool;
 | BoolGtF(_loc, a, b) ->
   typer a TFloat32;
   typer b TFloat32;
   body.t <- TBool;
 | BoolGtF64(_loc, a, b) ->
   typer a TFloat64;
   typer b TFloat64;
   body.t <- TBool;  

 | BoolLtE(_loc, a, b) ->
   typer a TInt;
   typer b TInt;
   body.t <- TBool;
 | BoolLtE32(_loc, a, b) ->
   typer a TInt32;
   typer b TInt32;
   body.t <- TBool;
 | BoolLtE64(_loc, a, b) ->
   typer a TInt64;
   typer b TInt64;
   body.t <- TBool;
 | BoolLtEF(_loc, a, b) ->
   typer a TFloat32;
   typer b TFloat32;
   body.t <- TBool;
 | BoolLtEF64(_loc, a, b) ->
   typer a TFloat64;
   typer b TFloat64;
   body.t <- TBool;  


 | BoolGtE(_loc, a, b) ->
   typer a TInt;
   typer b TInt;
   body.t <- TBool;
 | BoolGtE32(_loc, a, b) ->
   typer a TInt32;
   typer b TInt32;
   body.t <- TBool;
 | BoolGtE64(_loc, a, b) ->
   typer a TInt64;
   typer b TInt64;
   body.t <- TBool;
 | BoolGtEF(_loc, a, b) ->
   typer a TFloat32;
   typer b TFloat32;
   body.t <- TBool;
 | BoolGtEF64(_loc, a, b) ->
   typer a TFloat64;
   typer b TFloat64;
   body.t <- TBool;  

 | Ife (_loc, cond, cons1, cons2) ->
   (typer cond  TBool;
    typer cons1 t;	
    typer cons2 t;
    if cons1.t = TUnknown || cons1.t = TVec TUnknown then
      typer cons1 cons2.t;
    if cons2.t = TUnknown || cons2.t = TVec TUnknown then
      typer cons2 cons1.t;
    (match cons1.t,cons2.t with
    | TUnknown, TUnknown -> ()
    | TUnknown, _ -> cons1.t <- cons2.t
    | _, TUnknown -> cons2.t <- cons1.t
    | tt1, tt2 -> 
      if tt1 != tt2 then 
	( (* assert false *) (); raise (TypeError (tt1, tt2, cons2.loc)))
      else ());
    body.t <- cons1.t;
   )
 | If (_loc, cond, cons1) ->
   (typer cond  TBool;
    typer cons1 TUnknown;	
    body.t <- cons1.t;
   )
  | DoLoop (l, id, min, max, body) ->
    (match id.e with
    | Id (_loc,s)  ->
      (incr arg_idx;
       Hashtbl.add !current_args (string_of_ident s) 
	 {n = !arg_idx; var_type = TInt;
	  is_mutable = false;
	  read_only = false;
	  write_only = false;
	  is_global = false;
	 }) 
    | CastId (tt, Id(_loc,s)) ->
      (incr arg_idx;
       Hashtbl.add !current_args (string_of_ident s) 
	 {n = !arg_idx; var_type = TInt;
	  is_mutable = false;
	  read_only = false;
	  write_only = false;
	  is_global = false;})
    | _  ->  ()
    );
    typer id TInt;
    typer min TInt;
    typer max TInt;
    typer body t;
    body.t <- TUnit;
  | While (l, cond, body) ->
    typer cond TBool;
    typer body TUnit;
    body.t <- TUnit;
  | App (l, e1, e2) -> 
    let t = typer_app e1 e2 t in
    body.t <- t
  | Open (l, id, e) ->
	let rec aux = function
		| IdAcc (l,a,b) -> aux a; aux b
		| IdUid (l,s) -> open_module s l
		| _ -> (* assert false *) ()
		in
		aux id;
 (*   open_module (string_of_ident id) l; *)
    typer e TUnknown;
		let rec aux = function
		| IdAcc (l,a,b) -> aux a; aux b
		| IdUid (l,s) -> close_module s
		| _ -> (* assert false *) ()
		in
		aux id;
(*		close_module (string_of_ident id); *)
    body.t <- e.t
  | Noop -> ()
  | BoolAnd (l, e1, e2)  | BoolOr (l, e1, e2) ->
    typer e1 TBool;
    typer e2 TBool;
    body.t <- TBool
  | Ref (l,i) ->
    typer i t;
    body.t <- i.t
  | ModuleAccess(l,s,e) ->
    open_module s l;
    typer e TUnknown;
		close_module (s);
    body.t <- e.t;
  | _ -> (* assert false *) ()

and parse_app a =
	 my_eprintf (Printf.sprintf "(* val2 parse_app %s *)\n%!" (k_expr_to_string a.e));
  match a.e with
	| App (_loc, e1, e2::[]) ->
		let res = ref [] in
		let rec aux app =
			match app.e with
			| 	Id (_loc, s) ->
					(try 
	 				 let intr = Hashtbl.find !intrinsics_fun (string_of_ident s) in
      		 <:expr< intrinsics $ExStr(_loc, intr.cuda_val)$ $ExStr(_loc, intr.opencl_val)$>> 
						with Not_found -> 
						parse_body2 e1 false;)
			| App (_loc, e3, e4::[]) ->
				let e = aux e3 in
				res := <:expr< ($parse_body2 e4 false$)>> :: !res;
				e
			| ModuleAccess (_loc, s, e3) ->
				 open_module s _loc; 
    				let e = aux e3 in
    				close_module s;
    				e
			 | _ 	-> assert false; 
			in 
			let intr = aux e1 in
			res := (parse_body2 e2 false) :: !res ; 
			(match !res with
			| [] -> assert false 
			| t::[] -> 
				<:expr< app $intr$ [| ($t$) |]>>
			| t::q ->
				<:expr< app $intr$ [| $exSem_of_list (List.rev !res)$ |]>>)
	| _ -> parse_body2 a false
			
(* 		(match e1.e with                                                                                     *)
(*       	|	Id (_loc, s) ->                                                                               *)
(* 					(try                                                                                           *)
(*        		let intr =                                                                                    *)
(* 	 				 let intr = Hashtbl.find !intrinsics_fun (string_of_ident s) in                               *)
(*       		 <:expr< intrinsics $ExStr(_loc, intr.cuda_val)$ $ExStr(_loc, intr.opencl_val)$>> in    			 *)
(* 					let res = ref [] in                                                                            *)
(* 					let rec aux e2 =                                                                               *)
(* 						match e2 with                                                                                *)
(* 						| t::q ->                                                                                    *)
(* 							(match t.e with                                                                            *)
(* 							| App (l, e1, e2) ->                                                                       *)
(* 								res := (parse_body2 e1 false) :: !res ;                                                  *)
(* 								aux e2;                                                                                  *)
(* 							| _ ->                                                                                     *)
(* 							res := (parse_body2 t false):: !res )                                                      *)
(* 						| [] -> ()                                                                                   *)
(* 					in                                                                                             *)
(* 					aux (e2);                                                                                      *)
(* 	  			<:expr< app $intr$ [ $exSem_of_list (List.rev !res)$ ]>>                                       *)
					
(* 					with Not_found ->                                                                              *)
(* 						parse_body2 e1 false;                                                                        *)
(* (*						my_eprintf "unbound value %s\n" (string_of_ident s); (* assert false *) ())) *) )               *)
(* 				| App (_loc, _, _) ->                                                                            *)
(* 					le                                                                                             *)
(* 					 <:expr< app $parse_app e1$ $parse_body2 (List.hd e2) false$>>	                               *)
(* 				| ModuleAccess (_loc, s, e) ->                                                                   *)
(* 					 open_module s _loc;                                                                           *)
(*     				let e = parse_app e in                                                                       *)
(*     				close_module s;                                                                              *)
(*     				e                                                                                            *)
(* 				| _ -> parse_body2 e1 false	                                                                    *)
(* 				| _ -> my_eprintf "(* app %s *)\n%!" (k_expr_to_string e1.e); (* assert false *) ())                  *)
(*   | Id (_loc,s) ->                                                                                       *)
(*     (try                                                                                                 *)
(*        let intr =                                                                                        *)
(* 	 Hashtbl.find !intrinsics_fun (string_of_ident s) in                                                   *)
(*        <:expr< intrinsics $ExStr(_loc, intr.cuda_val)$ $ExStr(_loc, intr.opencl_val)$>>                  *)
(*      with Not_found -> my_eprintf "unbound value %s\n" (string_of_ident s); (* assert false *) ())            *)
(*   | ModuleAccess (_loc, s, e) ->                                                                         *)
(*     open_module s _loc;                                                                                  *)
(*     let e = parse_app e in                                                                               *)
(*     close_module s;                                                                                      *)
(*     e                                                                                                    *)
(* 	| _ -> parse_body2 a false                                                                             *)
(*   | _ ->                                                                                                 *)
(*     my_eprintf "(* app %s *)\n%!" (k_expr_to_string a.e);                                            *)
(*     (* assert false *) ()                                                                                         *)

 
and expr_of_app t _loc gen_var y =
  match t with
  | TApp (t1,((TApp (t2,t3)) as tt)) ->
    expr_of_app tt _loc gen_var y
  | TApp (t1,t2) -> 
    (match t2 with
    | TInt -> <:expr<(new_int_var $ExInt(_loc,string_of_int gen_var.n)$)>>, (parse_body2 y false)
    | TInt32 -> <:expr<(new_int_var $ExInt(_loc,string_of_int gen_var.n)$)>>, (parse_body2 y false)
    | TInt64 -> <:expr<(new_int_var $ExInt(_loc,string_of_int gen_var.n)$)>>, (parse_body2 y false)
    | TFloat32 -> <:expr<(new_float_var $ExInt(_loc,string_of_int gen_var.n)$)>>,(parse_body2 y false)
    | TFloat64 -> <:expr<(new_double_var $ExInt(_loc,string_of_int gen_var.n)$)>>, (parse_body2 y false)
    | _  -> (*assert false *) (); failwith "unknown var type")
  | _ -> assert false 

and parse_body2 body bool = 
  (*  let _loc = Loc.ghost in*)
  let rec aux ?return_bool:(r=false) body =
    my_eprintf (Printf.sprintf "(* val2 %s *)\n%!" (k_expr_to_string body.e));
    match body.e with
    | Bind (_loc, var,y, z, is_mutable)  ->             
      (
	(
	  match var.e with 
	  | Id (_loc, s)  -> 
            (let gen_var = 
	       try (Hashtbl.find !current_args (string_of_ident s)) 
	       with _ -> assert false in
             let ex1,ex2 = 
               match var.t with
               | TInt -> <:expr<(new_int_var $ExInt(_loc,string_of_int gen_var.n)$)>>, (aux y)
               | TInt32 -> <:expr<(new_int_var $ExInt(_loc,string_of_int gen_var.n)$)>>, (aux y)
               | TInt64 -> <:expr<(new_int_var $ExInt(_loc,string_of_int gen_var.n)$)>>, (aux y)
               | TFloat32 -> <:expr<(new_float_var $ExInt(_loc,string_of_int gen_var.n)$)>>,(aux y)
               | TFloat64 -> <:expr<(new_double_var $ExInt(_loc,string_of_int gen_var.n)$)>>,(aux y)
	       			 | TApp _ -> expr_of_app var.t _loc gen_var y
               | _  -> (raise (TypeError (TUnknown, gen_var.var_type , _loc));
												(* assert false *) (); failwith "unknown var type")
             in
             arg_list := <:expr<(spoc_declare $ex1$)>>:: !arg_list)
	  | _  ->  failwith "strange binding" );
	(let var_ = parse_body2 var false in
	 let y = aux y in
	 let z = aux z in
	 let res = <:expr< seq (spoc_set $var_$ $y$) $z$>>
	 in remove_int_var var;
	 res))
 
    | Plus (_loc, a,b) -> body.t <- TInt; 
      let p1 = (parse_int2 a TInt) 
      and p2 = (parse_int2 b TInt) in
      if not r then 
	return_type := TInt;
      ( <:expr<spoc_plus $p1$ $p2$>>) 
    | Plus32 (_loc, a,b) -> body.t <- TInt32; 
      let p1 = (parse_int2 a TInt32) 
      and p2 = (parse_int2 b TInt32) in
      if not r then 
	return_type := TInt32;
      ( <:expr<spoc_plus $p1$ $p2$>>) 
    | Plus64 (_loc, a,b) -> body.t <- TInt64; 
      let p1 = (parse_int2 a TInt64) 
      and p2 = (parse_int2 b TInt64) in
      if not r then 
	return_type := TInt64;
      ( <:expr<spoc_plus $p1$ $p2$>>) 
    | PlusF (_loc, a,b) -> 
      let p1 = (parse_float2 a TFloat32) 
      and p2 = (parse_float2 b TFloat32) in
      if not r then 
	return_type := TFloat32;
      ( <:expr<spoc_plus_float $p1$ $p2$>>) 
    | PlusF32 (_loc, a,b) -> 
      let p1 = (parse_float2 a TFloat32) 
      and p2 = (parse_float2 b TFloat32) in
      if not r then 
	return_type := TFloat32;
      ( <:expr<spoc_plus_float $p1$ $p2$>>) 
    | PlusF64 (_loc, a,b) -> 
      let p1 = (parse_float2 a TFloat64) 
      and p2 = (parse_float2 b TFloat64) in
      if not r then 
	return_type := TFloat64;
      ( <:expr<spoc_plus_float $p1$ $p2$>>) 
    | Min (_loc, a,b) -> body.t <- TInt; 
      let p1 = (parse_int2 a TInt) 
      and p2 = (parse_int2 b TInt) in
      if not r then 
	return_type := TInt;
      ( <:expr<spoc_min $p1$ $p2$>>) 
    | Min32 (_loc, a,b) -> body.t <- TInt32; 
      ( <:expr<spoc_min $(parse_int2 a TInt32)$ $(parse_int2 b TInt32)$>>)
    | Min64 (_loc, a,b) -> body.t <- TInt64; 
      ( <:expr<spoc_min $(parse_int2 a TInt64)$ $(parse_int2 b TInt64)$>>)
    | MinF (_loc, a,b) -> 
      ( <:expr<spoc_min_float $(parse_float2 a TFloat32)$ $(parse_float2 b TFloat32)$>>)
    | MinF32 (_loc, a,b) -> 
      ( <:expr<spoc_min_float $(parse_float2 a TFloat32)$ $(parse_float2 b TFloat32)$>>)
    | MinF64 (_loc, a,b) -> 
      ( <:expr<spoc_min_float $(parse_float2 a TFloat64)$ $(parse_float2 b TFloat64)$>>)

    | Mul (_loc, a,b) -> body.t <- TInt; 
      let p1 = (parse_int2 a TInt) 
      and p2 = (parse_int2 b TInt) in
      if not r then 
	return_type := TInt;
      ( <:expr<spoc_mul $p1$ $p2$>>) 
    | Mul32 (_loc, a,b) -> body.t <- TInt32; 
      ( <:expr<spoc_mul $(parse_int2 a TInt32)$ $(parse_int2 b TInt32)$>>)
    | Mul64 (_loc, a,b) -> body.t <- TInt64; 
      ( <:expr<spoc_mul $(parse_int2 a TInt64)$ $(parse_int2 b TInt64)$>>)
    | MulF (_loc, a,b) -> 
      ( <:expr<spoc_mul_float $(parse_float2 a TFloat32)$ $(parse_float2 b TFloat32)$>>)
    | MulF32 (_loc, a,b) -> 
      ( <:expr<spoc_mul_float $(parse_float2 a TFloat32)$ $(parse_float2 b TFloat32)$>>)
    | MulF64 (_loc, a,b) -> 
      ( <:expr<spoc_mul_float $(parse_float2 a TFloat64)$ $(parse_float2 b TFloat64)$>>)

    | Div (_loc, a,b) -> body.t <- TInt; 
      let p1 = (parse_int2 a TInt) 
      and p2 = (parse_int2 b TInt) in
      if not r then 
	(return_type := TInt;
	);
      ( <:expr<spoc_div $p1$ $p2$>>) 
    | Div32 (_loc, a,b) -> body.t <- TInt32; 
      ( <:expr<spoc_div $(parse_int2 a TInt32)$ $(parse_int2 b TInt32)$>>)
    | Div64 (_loc, a,b) -> body.t <- TInt64; 
      ( <:expr<spoc_div $(parse_int2 a TInt64)$ $(parse_int2 b TInt64)$>>)
    | DivF (_loc, a,b) -> 
      ( <:expr<spoc_div_float $(parse_float2 a TFloat32)$ $(parse_float2 b TFloat32)$>>)
    | DivF32 (_loc, a,b) -> 
      ( <:expr<spoc_div_float $(parse_float2 a TFloat32)$ $(parse_float2 b TFloat32)$>>)
    | DivF64 (_loc, a,b) -> 
      ( <:expr<spoc_div_float $(parse_float2 a TFloat64)$ $(parse_float2 b TFloat64)$>>)

    | Mod (_loc, a,b) -> body.t <- TInt; 
      let p1 = (parse_int2 a TInt) 
      and p2 = (parse_int2 b TInt) in
      if not r then 
	return_type := TInt;
      ( <:expr<spoc_mod $p1$ $p2$>>) 

    | Id (_loc,s)  -> 
      (try 
	 let var = 
	   (Hashtbl.find !current_args (string_of_ident s))  in
	 ( match var.var_type with 
	 | TUnit -> body.t <- TUnit; <:expr< Unit>>
	 | _ -> 
	   body.t <- var.var_type;
	   (* return_type := var.var_type; HERE *)
	   if var.is_global then
	     match var.var_type with
	     | TFloat32 ->
		 <:expr<global_float_var (fun () -> $ExId(_loc,s)$)>>
	     | TInt -> <:expr<global_int_var (fun () -> $ExId(_loc,s)$)>>
	     | _ -> assert false 
	   else
	     <:expr<var  $ExInt(_loc, string_of_int var.n)$>>)
       with _  ->  
	 try 
	   let c_const = (Hashtbl.find !intrinsics_const (string_of_ident s))  in
	   if body.t <> c_const.typ then
	     if body.t = TUnknown then
	       body.t <- c_const.typ
	     else
	       ((* assert false *) (); raise (TypeError (c_const.typ, body.t, _loc)));
	   (* return_type := c_const.typ; HERE *)
	   <:expr<intrinsics $ExStr(_loc, c_const.cuda_val)$ $ExStr(_loc, c_const.opencl_val)$>>
	 with _ -> 
		(try 
       let intr = 
	 Hashtbl.find !intrinsics_fun (string_of_ident s) in
       <:expr< intrinsics $ExStr(_loc, intr.cuda_val)$ $ExStr(_loc, intr.opencl_val)$>>
     with Not_found -> raise (Unbound_value ((string_of_ident s), _loc))));
    | CastId (t, Id(_loc,s))  -> 
      let var = (Hashtbl.find !current_args (string_of_ident s)) in
      (match t with 
      | TFloat64 -> 
       	<:expr<double_var  $ExInt(_loc, string_of_int var.n)$>>
      | _ -> assert false )
    | Int (_loc, i)  -> <:expr<spoc_int $ExInt(_loc, i)$>>
    | Int32 (_loc, i)  -> <:expr<spoc_int32 $ExInt32(_loc, i)$>>
    | Int64 (_loc, i)  -> <:expr<spoc_int64 $ExInt64(_loc, i)$>>
    | Float (_loc, f)  -> <:expr<spoc_float $ExFlo(_loc, f)$>>
    | Float32 (_loc, f) -> <:expr<spoc_float $ExFlo(_loc, f)$>>
    | Float64 (_loc, f) -> <:expr<spoc_double $ExFlo(_loc, f)$>>
    | Seq (_loc, x, y) -> 
      (match y.e with
      | Seq _ ->
	let x = parse_body2 x false in
	let y = parse_body2 y false in
	  <:expr<seq $x$ $y$>>
      | _ -> 
	let e1 = parse_body2 x false in
	let e2 = aux y
	in  <:expr<seq $e1$ $e2$>> 
      )
    | End (_loc, x)  -> 
      let res = <:expr< $aux x$>> 
      in 
      (* return_type := x.t; HERE *)
      <:expr<$res$>> 
    | VecSet (_loc, vector, value)  -> 
      let gen_value = aux (~return_bool:true) value in
      let gen_value = 
	match vector.t, value.e with
	| TInt, (Int _) -> <:expr<( $gen_value$)>>
	| TInt32, (Int32 _) -> <:expr<( $gen_value$)>> 
	| TInt64, (Int64 _) -> <:expr<( $gen_value$)>> 
	| TFloat, (Float _) -> <:expr<( $gen_value$)>> 
	| TFloat32, (Float32 _) -> <:expr<( $gen_value$)>> 
	| TFloat64, (Float64 _) -> <:expr<( $gen_value$)>> 
	| _ -> gen_value
      in
      let v = aux  (~return_bool:true) vector in
      let e = <:expr<set_vect_var $v$ $gen_value$>> in
      return_type := TUnit; 
      e
    | VecGet(_loc, vector, index)  -> 
      let e =
			<:expr<get_vec $aux vector$ $parse_int2 index TInt$>> in
      (match vector.t with
      | TVec ty->
	();
	(*if not r then 
	  return_type := ty; *)
      | _ ->
	(* assert false *) ());
      e
    | BoolOr(_loc, a, b) ->
      <:expr< b_or $aux a$ $aux b$>>
    | BoolAnd(_loc, a, b) ->
      <:expr< b_and $aux a$ $aux b$>>
    | BoolEq(_loc, a, b) ->
      <:expr< equals $aux a$ $aux b$>>
    | BoolEq32(_loc, a, b) ->
      <:expr< equals32 $aux a$ $aux b$>>
    | BoolEq64(_loc, a, b) ->
      <:expr< equals64 $aux a$ $aux b$>>
    | BoolEqF(_loc, a, b) ->
      <:expr< equalsF $aux a$ $aux b$>>
    | BoolEqF64(_loc, a, b) ->
      <:expr< equalsF64 $aux a$ $aux b$>>
    | BoolLt(_loc, a, b) ->
      <:expr< lt $aux a$ $aux b$>>
    | BoolLt32(_loc, a, b) ->
      <:expr< lt32 $aux a$ $aux b$>>
    | BoolLt64(_loc, a, b) ->
      <:expr< lt64 $aux a$ $aux b$>>
    | BoolLtF(_loc, a, b) ->
      <:expr< ltF $aux a$ $aux b$>>
    | BoolLtF64(_loc, a, b) ->
      <:expr< ltF64 $aux a$ $aux b$>>

    | BoolGt(_loc, a, b) ->
      <:expr< gt $aux a$ $aux b$>>
    | BoolGt32(_loc, a, b) ->
      <:expr< gt32 $aux a$ $aux b$>>
    | BoolGt64(_loc, a, b) ->
      <:expr< gt64 $aux a$ $aux b$>>
    | BoolGtF(_loc, a, b) ->
      <:expr< gtF $aux a$ $aux b$>>
    | BoolGtF64(_loc, a, b) ->
      <:expr< gtF64 $aux a$ $aux b$>>

    | BoolLtE(_loc, a, b) ->
      <:expr< lte $aux a$ $aux b$>>
    | BoolLtE32(_loc, a, b) ->
      <:expr< lte32 $aux a$ $aux b$>>
    | BoolLtE64(_loc, a, b) ->
      <:expr< lte64 $aux a$ $aux b$>>
    | BoolLtEF(_loc, a, b) ->
      <:expr< lteF $aux a$ $aux b$>>
    | BoolLtEF64(_loc, a, b) ->
      <:expr< lteF64 $aux a$ $aux b$>>

    | BoolGtE(_loc, a, b) ->
      <:expr< gte $aux a$ $aux b$>>
    | BoolGtE32(_loc, a, b) ->
      <:expr< gte32 $aux a$ $aux b$>>
    | BoolGtE64(_loc, a, b) ->
      <:expr< gte64 $aux a$ $aux b$>>
    | BoolGtEF(_loc, a, b) ->
      <:expr< gteF $aux a$ $aux b$>>
    | BoolGtEF64(_loc, a, b) ->
      <:expr< gteF64 $aux a$ $aux b$>>
    | Ife (_loc, cond, cons1, cons2) ->
      (<:expr< spoc_ife $aux cond$ $aux cons1$ $aux cons2$>>)
    | If (_loc, cond, cons1) ->
      (<:expr< spoc_if $aux cond$ $aux cons1$>>)
    | DoLoop (_loc, id, min, max, body) ->
      (<:expr<spoc_do $aux id$ $aux min$ $aux max$ $aux body$>>)
    | While (_loc, cond, body) ->
      let cond = aux cond in
      let body = aux body in
      (<:expr<spoc_while $cond$ $body$>>)
    | Seqs exprs ->
      let _loc = body.loc in
      let rec aux = function
	| [] -> <:expr< Unit>>
	| t::[] -> 
	  let e2 = parse_body2 t false in
	  e2
	| t::q -> 
	  let q = aux q in
	  let e2 =  (parse_body2 t true) in
	  <:expr< (seq $e2$ $q$)>>
      in aux exprs
    | App (_loc, e1, e2) ->
      let e = <:expr< $parse_app body$>> in
      e
    | Open (_loc, id, e) ->
      let rec aux2 = function
		| IdAcc (l,a,b) -> aux2 a; aux2 b
		| IdUid (l,s) -> open_module s l
		| _ -> (* assert false *) ()
		in
		aux2 id;
      let ex = <:expr< $aux e$>> in
			let rec aux2 = function
		| IdAcc (l,a,b) -> aux2 a; aux2 b
		| IdUid (l,s) -> close_module s
		| _ -> (* assert false *) ()
		in
		aux2 id;
      (* return_type := e.t; HERE *)
      ex
    | ModuleAccess (_loc, s, e) ->
      open_module s _loc;
      let ex = <:expr< $aux e$>> in
      close_module s;
      ex
    | Noop ->
      let _loc = body.loc in
      <:expr< spoc_unit () >>
    | Acc (_loc, e1, e2) ->
      let e1 = parse_body2 e1 false 
      and e2 = parse_body2 e2 false
      in
      if not r then 
	return_type := TUnit;
      <:expr< spoc_acc $e1$ $e2$>>
    | Ref (_loc, {t=_; e=Id(_loc2, s);loc=_}) ->
      let var = 
	Hashtbl.find !current_args (string_of_ident s)  in
      body.t <- var.var_type;
      (* return_type := var.var_type; HERE *)
      if var.is_global then
	match var.var_type with
	| TFloat32 ->
	  <:expr<global_float_var (fun _ -> ! $ExId(_loc,s)$)>>
	| TInt -> <:expr<global_int_var (fun _ -> ! $ExId(_loc,s)$)>>
	| _ -> assert false 
      else
	assert false 
	  
    | _ -> (* assert false *) (); failwith "unimplemented yet"
  in
  let _loc = body.loc in
  if bool then 
		(
	    my_eprintf (Printf.sprintf"(* val2 %s *)\n%!" (k_expr_to_string body.e));
    match body.e with
    | Bind (_loc, var,y, z, is_mutable)  -> 
      (
	(match var.e with 
	| Id (_loc, s) ->
	  (let gen_var = (Hashtbl.find !current_args (string_of_ident s)) in
           let ex1,ex2 = 
             match var.t with
             | TInt -> <:expr<(new_int_var $ExInt(_loc,string_of_int gen_var.n)$)>>, (aux y)
             | TInt32 -> <:expr<(new_int_var $ExInt(_loc,string_of_int gen_var.n)$)>>,(aux y)
             | TInt64 -> <:expr<(new_int_var $ExInt(_loc,string_of_int gen_var.n)$)>>, (aux y)
             | TFloat32 -> <:expr<(new_float_var $ExInt(_loc,string_of_int gen_var.n)$)>>,(aux y)
             | TFloat64 -> <:expr<(new_double_var $ExInt(_loc,string_of_int gen_var.n)$)>>,(aux y)
	     | TUnit -> <:expr<Unit>>,aux y; (*<:expr<(new_unit_var $ExInt(_loc,string_of_int gen_var.n)$)>>,(aux y)*)
             | _  -> (* assert false *) (); failwith "unknown var type"
           in
           match var.t with
	   |  TUnit ->
	     arg_list := <:expr<Seq ($ex1$, $ex2$)>>::!arg_list
	   | _ -> arg_list := <:expr<spoc_set $ex1$ $ex2$>>:: !arg_list);
	| CastId (_,Id(_loc,s)) -> 
          (let gen_var = (Hashtbl.find !current_args (string_of_ident s)) in
           let ex1,ex2 = 
             match var.t with
             | TInt -> <:expr<(new_int_var $ExInt(_loc,string_of_int gen_var.n)$)>>,(aux y)
             | TInt32 -> <:expr<(new_int_var $ExInt(_loc,string_of_int gen_var.n)$)>>,(aux y)
             | TInt64 -> <:expr<(new_int_var $ExInt(_loc,string_of_int gen_var.n)$)>>, (aux y)
             | TFloat32 -> <:expr<(new_double_var $ExInt(_loc,string_of_int gen_var.n)$)>>,(aux y)
             | TFloat64 -> <:expr<(new_double_var $ExInt(_loc,string_of_int gen_var.n)$)>>,(aux y)
             | _  -> (* assert false *) (); failwith "unknown var type"
           in
           arg_list := <:expr<spoc_set $ex1$ $ex2$>>:: !arg_list);
	| _  ->  (* assert false *) (); failwith "strange vector" );
	let res = <:expr<$parse_body2 z true$>>
	in remove_int_var var;
	res)
    | Seq (a,b,c)  -> aux body
    | _  -> 
      let e = {t=body.t; e =End(_loc, body); loc = _loc} in 
      match body.t with 
      | TUnit ->
	let res = aux e in
	return_type := TUnit;
	<:expr< $res$ >>
      |_ -> 
	<:expr<spoc_return $aux e$>>
	)
  else
    aux body

      
      
let parse_args params body= 
  let _loc = Loc.ghost in
  let rec aux acc = function
    | []  ->  acc
    | t::q -> aux <:expr<fun $t$ -> $acc$>> q
  in aux body
  (List.rev params)
  



let new_arg_of_patt p =  
(*  let _loc = Loc.ghost in *)
  match p with
  | <:patt< $lid:x$ >> ->  let i = !arg_idx in     incr arg_idx;

                           Hashtbl.add !current_args x {n=i; var_type=TUnknown;
							is_mutable = false;
							read_only = false;
							write_only = false;
							is_global = false;};
  | _  -> failwith "error new_arg_of_patt"

let gen_arg_from_patt p = 
  match p with
  | PaId (_loc, x)   -> <:expr<(spoc_id $ExId(_loc,x)$)>>
  | _  -> failwith "error gen_arg_from_patt"
    

let gen_arg_from_patt2 p =  
	  (*  let _loc = Loc.ghost in*)
  match p with
  | PaId (_loc, IdLid (_loc2, x) ) -> 
      (*	  <:patt< $lid:x$ >> ->*)
    (let var = (Hashtbl.find !current_args x) in   
     match var.var_type with
     | TUnknown  -> <:expr<(new_unknown_var $ExInt(_loc2,  string_of_int var.n)$)>>(* failwith "Forbidden unknown argument type in kernel declaration"*)
     | TInt | TInt32 | TInt64 ->                                                 
       <:expr<(new_int_var $ExInt(_loc2,  string_of_int var.n)$)>>
     | TFloat32 | TFloat -> 
       <:expr<(new_float_var $ExInt(_loc2,  string_of_int var.n)$)>>
		 | TFloat64  -> 
       <:expr<(new_float64_var $ExInt(_loc2,  string_of_int var.n)$)>>
     | TVec k -> 
       (match k with
       | TInt | TInt32 | TInt64  ->   <:expr<(new_int_vec_var $ExInt(_loc2,  string_of_int var.n)$)>>
       | TFloat | TFloat32 ->   <:expr<(new_float_vec_var $ExInt(_loc2,  string_of_int var.n)$)>>
       | TFloat64 ->   <:expr<(new_double_vec_var $ExInt(_loc2,  string_of_int var.n)$)>>
       | _  -> failwith "Forbidden vector  type in kernel declaration")
     | _ -> (*(* assert false *) ();*) failwith "unimplemented yet"
    )
  | _  -> failwith "error gen_arg_form_patt2"




let patt_is_vector p =
  match p with
  | PaId (_loc, IdLid (_loc2, x) ) -> 
    (*	  <:patt< $lid:x$ >> ->*)
    let var = (Hashtbl.find !current_args x) in   
    match var.var_type with
    | TVec _ ->true
    | _ -> false
  | _ -> assert false 


let gen_arg_from_patt3 p =  
  (*  let _loc = Loc.ghost in*)
  match p with
  | PaId (_loc, IdLid (_loc2, x) ) -> 
    (*	  <:patt< $lid:x$ >> ->*)
    (let var = (Hashtbl.find !current_args x) in   
     let s = <:ident< $lid:("spoc_var"^string_of_int var.n)$>> in
     let v = <:patt< $id:s$ >> in
     let e1, e2, e3, e4, e5 =
       match var.var_type with
       | TUnknown  -> 
	 <:ctyp< 'a>>, 
	 <:expr< Spoc.Kernel.abs>>,
	 <:ctyp< 'a>>,
	 IdAcc(_loc, 
	       IdUid(_loc, "Spoc"), 
	       IdAcc(_loc, 
		     IdUid (_loc, "Kernel"), 
		     IdLid(_loc, "abs"))),
	 <:ctyp< 'a>>
       (* failwith "Forbidden unknown argument type in kernel declaration"*)
       | TInt -> 
	 <:ctyp< int>>, 
	 <:expr< Spoc.Kernel.Int32>>,
	 <:ctyp< int>>,
	 IdAcc(_loc, 
	       IdUid(_loc, "Spoc"), 
	       IdAcc(_loc, 
		     IdUid (_loc, "Kernel"), 
		     IdUid(_loc, "Int32"))),
	 <:ctyp< int>>
       | TInt32 ->  
	 <:ctyp< int32>>, 
	 <:expr< Spoc.Kernel.Int32>>,
	 <:ctyp< int32>>,
	 IdAcc(_loc, 
	       IdUid(_loc, "Spoc"), 
	       IdAcc(_loc, 
		     IdUid (_loc, "Kernel"), 
		     IdUid(_loc, "Int32"))),
	 <:ctyp< int32>>
       | TInt64 ->   
	 <:ctyp< int64>>, 
	 <:expr< Spoc.Kernel.Int64>>,
	 <:ctyp< int32>>,
	 IdAcc(_loc, 
	       IdUid(_loc, "Spoc"), 
	       IdAcc(_loc, 
		     IdUid (_loc, "Kernel"), 
		     IdUid(_loc, "Int64"))),
	 <:ctyp< int64>>
   | TFloat | TFloat32 | TFloat64 ->  
	 <:ctyp< float>>,
	 <:expr< Spoc.Kernel.Float32>>,
	 <:ctyp< float>>,
	 IdAcc(_loc, 
	       IdUid(_loc, "Spoc"), 
	       IdAcc(_loc, 
		     IdUid (_loc, "Kernel"), 
		     IdUid(_loc, "Float32"))),
	 <:ctyp< float>>
       | TVec k -> 
	 (match k with
	 | TInt -> 
	   <:ctyp< (('spoc_a, 'spoc_b) Vector.vector)>>, 
	   <:expr< Spoc.Kernel.VInt32>>,
	   <:ctyp< ((int32, Bigarray.int32_elt) Vector.vector)>> ,
	   IdAcc(_loc, 
		 IdUid(_loc, "Spoc"), 
		 IdAcc(_loc, 
		       IdUid (_loc, "Kernel"), 
		       IdUid(_loc, "VInt32"))),
	   <:ctyp< Spoc.Vector.vint32>>
	 | TInt32 -> 
	   <:ctyp< (('spoc_c, 'spoc_d) Vector.vector)>>, 
	   <:expr< Spoc.Kernel.VInt32>>,
	   <:ctyp< ((int32, Bigarray.int32_elt) Vector.vector)>>,
	   IdAcc(_loc, 
		 IdUid(_loc, "Spoc"), 
		 IdAcc(_loc, 
		       IdUid (_loc, "Kernel"), 
		       IdUid(_loc, "VInt32"))),
	   <:ctyp< Spoc.Vector.vint32>>
	 | TInt64  -> 
	   <:ctyp< (('spoc_e, 'spoc_f) Vector.vector)>>, 
	   <:expr< Spoc.Kernel.VInt64>>,
	   <:ctyp< ((int64, Bigarray.int64_elt) Vector.vector)>>,
	   IdAcc(_loc, 
		 IdUid(_loc, "Spoc"), 
		 IdAcc(_loc, 
		       IdUid (_loc, "Kernel"), 
		       IdUid(_loc, "VInt64"))),
	   <:ctyp< Spoc.Vector.vint64>>
	 | TFloat -> 
	   <:ctyp< (('spoc_g, 'spoc_h) Vector.vector)>>, 
	   <:expr< Spoc.Kernel.VFloat32>>,
	   <:ctyp< ((float, Bigarray.float32_elt) Vector.vector)>>,
	   IdAcc(_loc, 
		 IdUid(_loc, "Spoc"), 
		 IdAcc(_loc, 
		       IdUid (_loc, "Kernel"), 
		       IdUid(_loc, "VFloat32"))),
	   <:ctyp< Spoc.Vector.vfloat32>>
	 | TFloat32 -> 
	   <:ctyp< (('spoc_i, 'spoc_j) Vector.vector)>>, 
	   <:expr< Spoc.Kernel.VFloat32>>,
	   <:ctyp< ((float, Bigarray.float32_elt) Vector.vector)>>,
	   IdAcc(_loc, 
		 IdUid(_loc, "Spoc"), 
		 IdAcc(_loc, 
		       IdUid (_loc, "Kernel"), 
		       IdUid(_loc, "VFloat32"))),
	   <:ctyp< Spoc.Vector.vfloat32>>
	 | TFloat64 ->  
	   <:ctyp< (('spoc_k, 'spoc_l) Vector.vector)>>, 
	   <:expr< Spoc.Kernel.VFloat64>>,
	   <:ctyp< ((float, Bigarray.float64_elt) Vector.vector)>>,
	   IdAcc(_loc, 
		 IdUid(_loc, "Spoc"), 
		 IdAcc(_loc, 
		       IdUid (_loc, "Kernel"), 
		       IdUid(_loc, "VFloat64"))),
	   <:ctyp< Spoc.Vector.vfloat64>>
	 | _  -> failwith "Forbidden vector  type in kernel declaration")
 | _ -> failwith "Forbidden vector  type in kernel declaration"; (* assert false *) (); failwith "unimplemented yet"
     in
     (*e*)
     match var.var_type with
     | TVec _ ->
       PaTyc (_loc, v, e1), 
       <:expr< $e2$ (Spoc.Kernel.relax_vector $id:s$)>>, 
       e3, 
       PaApp (_loc,
	      PaId(_loc, e4),
	      PaId(_loc,s)), 
       (if !new_kernel then
	   ( new_kernel := false;
	     <:expr< ($id:s$: $e3$)>>)
	else
	   <:expr< ( Spoc.Kernel.relax_vector $id:s$: $e3$)>>), e5
     | _ ->
       PaTyc (_loc, v, e1), 
       <:expr< $e2$ ($id:s$)>>, 
       e3, 
       PaApp (_loc,
	      PaId(_loc, e4),
	      PaId(_loc,s)), 
       <:expr< ($id:s$: $e3$)>> , e5
    (*<:expr< $v$:$e$ >>*)
    )
  | _  -> failwith "error gen_arg_form_patt2"




let rec float32_expr f = 
  let rec f32_typer f =
    (match f.t with
    | TUnknown -> f.t <- TFloat32
    | TFloat -> f.t <- TFloat32
    | TVec TFloat
    | TVec TFloat32 -> f.t <- TVec TFloat32
    | TFloat32-> ()
    | _ -> (* assert false *) (); raise (TypeError (TFloat32, f.t, f.loc)))
  in f32_typer f;
  
  (match f.e with
  | PlusF (l, a,b) 
  | PlusF32 (l, a,b) -> 
    f.e <- PlusF32 (l, float32_expr a, float32_expr b)
  | Float (l,s) 
  | Float32 (l,s) -> f.e<- Float32 (l, s)

  | VecGet (l,a,b) -> ()
  | Seq (l, a, b) -> f.e <- Seq (l,a, float32_expr b)
  | Id  (l, id) -> ()
  | _ -> (* assert false *) (); raise (TypeError (TFloat32, f.t, f.loc)));
  f

let rec float64_expr f = 
  let rec f64_typer f =
    (match f.t with
    | TUnknown -> ()
    | TFloat -> f.t <- TFloat64
    | TVec TFloat
    | TVec TFloat64 -> f.t <- TVec TFloat64
    | TFloat64-> ()
    | _ ->() ) 
  in f64_typer f;
  (match f.e with
  | PlusF (l, a,b) 
  | PlusF64 (l, a,b) -> 
    (
      f.e <- PlusF64 (l, float64_expr a, float64_expr b)
    )
  | MinF (l, a,b) 
  | MinF64 (l, a,b) -> 
    (
      f.e <- MinF64 (l, float64_expr a, float64_expr b)
    )

  | MulF (l, a,b) 
  | MulF64 (l, a,b) -> 
    (
      f.e <- MulF64 (l, float64_expr a, float64_expr b)
    )

  | DivF (l, a,b) 
  | DivF64 (l, a,b) -> 
    (
      f.e <- DivF64 (l, float64_expr a, float64_expr b)
    )
  | App (l, a, b) ->
    f.e <- App (l, float64_expr a, (List.map float64_expr b))
  | Float32 (l,s) 
  | Float64 (l,s) -> 
    (
      f.e<- Float64 (l, s))
  | Bind (l, e1, e2, e3, is_mutable) -> 
    f.e <- Bind (l, float64_expr e1, float64_expr e2, float64_expr e3, is_mutable)
  | VecGet (l,a,b) -> ()
  | VecSet (l,a,b) -> 
    f.e <- VecSet (l,float64_expr a, float64_expr b)
  | Seq (l, a, b) -> f.e <- Seq (l,a, float64_expr b)
  | Id  (l, id) -> f.e <- (CastId (TFloat64, Id  (l, id)))
  | Int _ | BoolEq _ | BoolEq32 _ | BoolEq64 _ -> () 
  | BoolLt _ | BoolLt32 _ | BoolLt64 _ 
  | Plus _ | Min _ | Mul _ | Div _ ->() 
  | BoolEqF (l,a,b) ->
    f.e <- BoolEqF64 (l, float64_expr a, float64_expr b)
  | BoolLt64 _-> ()
  | BoolLtF (l,a,b) -> 
    f.e <- BoolLtF64 (l, float64_expr a, float64_expr b)
  | Ife (l,cond,cons1,cons2) -> 
    f.e <- Ife (l, float64_expr cond, float64_expr cons1, float64_expr cons2)
  | If (l,cond,cons1) -> 
    f.e <- If (l, float64_expr cond, float64_expr cons1)
  | DoLoop (l, id, min, max, body) ->
    f.e <- DoLoop (l, id, min, max, 	float64_expr (body))
  | Open (l, m, e) -> 
    f.e <- Open (l, m, float64_expr e)
  | _ -> raise (TypeError (TFloat64, f.t, f.loc));
    ());
  f


let nb_ker = ref 0

let gen_kernel ()= () 
(*  let _loc = Loc.ghost in
  <:expr< () >> *)
    
    EXTEND Gram
    GLOBAL: str_item expr;
  expr: 
    [
      ["kern"; args = LIST1 k_patt; "->"; body = kernel_body ->
       arg_idx := 0;
       return_type := TUnknown;
       arg_list := [];
			 extensions := [ ex32 ];
       Hashtbl.clear !current_args;
       (*let n_body = *)
       (*ignore(<:expr<params 
         $List.fold_left
         (fun a b -> <:expr<concat $b$ $a$>>) 
         <:expr<empty_arg()>> 
         (List.rev (List.map gen_arg_from_patt args))$>>); *)
       List.iter new_arg_of_patt args;
       (try 
	  typer body TUnknown
	with
	| TypeError(expected, given, loc) -> 
	  (
	    Printf.eprintf "%s\n%!" ("\027[31m Type Error \027[00m : expecting : \027[33m"^
			 (ktyp_to_string expected)^"\027[00m but given : \027[33m"^
			 (ktyp_to_string given)^"\027[00m in position : "^(Loc.to_string loc)^"");
			exit 1;
			)
	| Unbound_value (value, loc) ->
		(Printf.eprintf "%s\n%!" ("\027[31m Unbound Value \027[00m : \027[33m"^
			 (value)^"\027[00m in position : "^(Loc.to_string loc)^"");
			exit 2;)
		
		);  
       let new_hash_args = Hashtbl.create (Hashtbl.length !current_args) in
       Hashtbl.iter (Hashtbl.add new_hash_args) !current_args;
       Hashtbl.clear !current_args;
       current_args := new_hash_args;  
       let gen_body = 
         <:expr< 
           (*spoc_fun_kernel 
             $n_body$*)
           $try parse_body body with
             | TypeError(expected, given, loc) -> 
	  (
	    Printf.eprintf "%s\n%!" ("\027[31m Type Error \027[00m : expecting : \027[33m"^
			 (ktyp_to_string expected)^"\027[00m but given : \027[33m"^
			 (ktyp_to_string given)^"\027[00m in position : "^(Loc.to_string loc)^"");
			exit 1;
			)
	| Unbound_value (value, loc) ->
		(Printf.eprintf "%s\n%!" ("\027[31m Unbound Value \027[00m : \027[33m"^
			 (value)^"\027[00m in position : "^(Loc.to_string loc)^"");
			exit 2;)
		$>>
       in
       (*print_int (List.length !arg_list);*)
       let b_body = 
         (try parse_body2 body true
          with 
          | TypeError(expected, given, loc) -> 
	  (
	    Printf.eprintf "%s\n%!" ("\027[31m Type Error \027[00m : expecting : \027[33m"^
			 (ktyp_to_string expected)^"\027[00m but given : \027[33m"^
			 (ktyp_to_string given)^"\027[00m in position : "^(Loc.to_string loc)^"");
			exit 1;
			)
	| Unbound_value (value, loc) ->
		(Printf.eprintf "%s\n%!" ("\027[31m Unbound Value \027[00m : \027[33m"^
			 (value)^"\027[00m in position : "^(Loc.to_string loc)^"");
			exit 2;))
       in
       let n_body2 = <:expr<params $List.fold_left 
         (fun a b -> <:expr<concat $b$ $a$>>) 
         <:expr<empty_arg()>> 
         (List.rev (List.map gen_arg_from_patt2 args))$>> in 
       let gen_body2 =  <:expr< 
         spoc_gen_kernel 
          $n_body2$
          $List.fold_left (fun a b -> 
            <:expr<spoc_local_env $b$ $a$>>)
          b_body
          !arg_list$>>

       in
       let gen_args = parse_args args gen_body
       in
       let ret =
	 incr arg_idx;
	 match !return_type with
         | TUnknown  -> <:expr<return_unknown ()>>(*failwith "Unknown return type"*)
         | TInt | TInt32 | TInt64 ->  <:expr<return_int $ExInt(Loc.ghost, string_of_int (!arg_idx))$>>
         | TVec TInt | TVec TInt32 | TVec TInt64 -> <:expr<return_vector_int $ExInt(Loc.ghost, string_of_int (!arg_idx))$>>
         | TFloat | TFloat32 ->  <:expr<return_float $ExInt(Loc.ghost, string_of_int (!arg_idx))$>>
	 | TFloat64 ->  <:expr<return_double $ExInt(Loc.ghost, string_of_int (!arg_idx))$>>
	 | TUnit  -> <:expr<return_unit ()>>
	 | TBool -> <:expr< return_bool $ExInt(Loc.ghost, string_of_int (!arg_idx))$>>
         | _  -> failwith "error ret"
       in
       let fst_ (a,b,c,d,e,f) = a
       and snd_ (a,b,c,d,e,f) = b
       and thd_ (a,b,c,d,e,f) = c 
       and fth_ (a,b,c,d,e,f) = d 
       and ffh_ (a,b,c,d,e,f) = e in
       let tup_args, list_args, class_legacy , list_to_args1, list_to_args2= 
	 let args_fst_list = 
	   (List.map fst_ (List.map gen_arg_from_patt3 args)) in
	 let args_com = paCom_of_list  args_fst_list in
	 
	
	 let args_list = 
		let l = (List.map snd_ (List.map gen_arg_from_patt3 args)) in
		exSem_of_list l
	  in 

	 let args_thd_list = 
	   (List.map thd_ (List.map gen_arg_from_patt3 args)) in
	 let args_typ = tySta_of_list args_thd_list in

	 let lta1 = 
	   paSem_of_list 
	   (List.map fth_ (List.map gen_arg_from_patt3 args)) in

	 let args_ffth_list = 	     
	   (new_kernel := true;
	    List.map ffh_ (List.map gen_arg_from_patt3 args)) in
	 let lta2 = 
	   exCom_of_list args_ffth_list in	 
	 
	 begin
	   if List.length args_fst_list = 1 then
	     List.hd args_fst_list
	   else
	     PaTup (_loc, args_com)
	 end, 
	   
	   begin
	     ExArr (_loc, args_list)
	   end,
	 
	     begin 
	       if List.length args_thd_list = 1 then
		 <:ctyp< $List.hd args_thd_list$,(('a,'b) Kernel.kernelArgs) array>>
	       else
		 <:ctyp< $TyTup (_loc, args_typ)$,(('a,'b) Kernel.kernelArgs) array>>
	     end,
		
	       begin
		 PaArr (_loc, lta1)
	       end,
		 
		 begin
		   if List.length args_ffth_list = 1 then
		     List.hd args_ffth_list
		   else
		     ExTup (_loc, lta2)
		 end
       in
       let nw = ExNew(_loc, IdLid(_loc, "kirc_class")) in 
       (*let (list_args: ctyp list) =
	 (List.map (fun (a,b,c,d,e,f) -> f)  (List.map gen_arg_from_patt3 args)) 
       in
       let ff tya tyb = TyArr (_loc, tyb, tya) in 
       let spoc_args = 
	 List.fold_right ff 
	   list_args  
	   (TyId (_loc, IdLid(_loc, "unit")))
       in*)
       let class_name = "kirc_class"^(string_of_int !nb_ker) in
       incr nb_ker;
       let has_vector = ref false in
       List.iter (fun a-> if patt_is_vector a then has_vector := true) args; 
				let extensions =  match !extensions with 
				| [] -> <:expr< [||]>>
				| t::[] -> <:expr< [|$t$|]>> 
				| _ -> <:expr<[|$exSem_of_list  !extensions$|]>>
       in
			if !has_vector then 
		 <:expr< let module M = 
		     struct 
		       let exec_fun $tup_args$ = Spoc.Kernel.exec $list_args$;;
		       class ['a, 'b] $lid:class_name$ = 
		       object (self)
			 inherit 
			   [$class_legacy$ ] 
			     Spoc.Kernel.spoc_kernel "kirc_kernel" "spoc_dummy"
			 method exec = exec_fun
			 method args_to_list = fun 
			   $tup_args$ -> $list_args$
			 method list_to_args = function
			 | $list_to_args1$ -> $list_to_args2$
			 | _ -> failwith "spoc_kernel_extension error"
		       end
		     end
		 in
		 let open Kirc in
		 (new M.$lid:class_name$, {
		   ml_kern = $gen_args$;
		   body = $gen_body2$;
		   ret_val = $ret$;
			 extensions = $extensions$;
			}
		 )>>
       else   
	 <:expr< let module M = 
		     struct
		       let exec_fun $tup_args$ = Spoc.Kernel.exec $list_args$;;
		       class ['a, 'b] $lid:class_name$ = 
		       object (self)
			 inherit 
			   [$class_legacy$ ] 
			     Spoc.Kernel.spoc_kernel "kirc_kernel" "spoc_dummy"
			 method exec = assert false 
			 method args_to_list = assert false 
			 method list_to_args = assert false 
		       end
		     end
		 in
		 let open Kirc in
		 (new M.$lid:class_name$, {
		   ml_kern = $gen_args$;
		   body = $gen_body2$;
		   ret_val = $ret$;
			 extensions = $extensions$;
		 })>>
      ]
    ];
  expr:
    [
      ["kfun"; args = LIST1 k_patt; "->"; body = kernel_body ->
      arg_idx := 0;
       return_type := TUnknown;
       arg_list := [];
       Hashtbl.clear !current_args;
       (*let n_body = *)
       (*ignore(<:expr<params 
         $List.fold_left
         (fun a b -> <:expr<concat $b$ $a$>>) 
         <:expr<empty_arg()>> 
         (List.rev (List.map gen_arg_from_patt args))$>>); *)
       List.iter new_arg_of_patt args;
       (try 
	  typer body TUnknown
	with
	| TypeError(expected, given, loc) -> 
	  (
	    failwith ("Type Error : expecting : "^
			 (ktyp_to_string expected)^" but given : "^
			 (ktyp_to_string given)^" in position : "^(Loc.to_string loc))));  
       let new_hash_args = Hashtbl.create (Hashtbl.length !current_args) in
       Hashtbl.iter (Hashtbl.add new_hash_args) !current_args;
       Hashtbl.clear !current_args;
       current_args := new_hash_args;  
       let gen_body = 
         <:expr< 
           (*spoc_fun_kernel 
             $n_body$*)
           $try parse_body body
             with 
             | TypeError(expected, given, loc) -> 
               (
		 failwith ("Type Error : expecting : "^
			      (ktyp_to_string expected)^" but given : "^
			      (ktyp_to_string given)^" in position : "^(Loc.to_string loc)))$>>
       in
       (*print_int (List.length !arg_list);*)
       let b_body = 
         (try parse_body2 body true
          with 
          | TypeError(expected, given, loc) -> 
            failwith ("Type Error : expecting : "^
			 (ktyp_to_string expected)^" but given : "^
			 (ktyp_to_string given)^" in position : "^(Loc.to_string loc)) )
       in
       let n_body2 = <:expr<params $List.fold_left 
         (fun a b -> <:expr<concat $b$ $a$>>) 
         <:expr<empty_arg()>> 
         (List.rev (List.map gen_arg_from_patt2 args))$>> in 
       let gen_body2 =  <:expr< 
         spoc_gen_kernel 
          $n_body2$
          $List.fold_left (fun a b -> 
            <:expr<spoc_local_env $b$ $a$>>)
          b_body
          !arg_list$>>
       in
       let gen_args = parse_args args gen_body
       in
       let ret =
	 incr arg_idx;
	 match !return_type with
         | TUnknown  -> <:expr<return_unknown ()>>(*failwith "Unknown return type"*)
         | TInt | TInt32 | TInt64 ->  <:expr<return_int $ExInt(Loc.ghost, string_of_int (!arg_idx))$>>
         | TVec TInt | TVec TInt32 | TVec TInt64 -> <:expr<return_vector_int $ExInt(Loc.ghost, string_of_int (!arg_idx))$>>
         | TFloat | TFloat32 ->  <:expr<return_float $ExInt(Loc.ghost, string_of_int (!arg_idx))$>>
	 | TFloat64 ->  <:expr<return_double $ExInt(Loc.ghost, string_of_int (!arg_idx))$>>
	 | TUnit  -> <:expr<return_unit ()>>
	 | TBool -> <:expr< return_bool $ExInt(Loc.ghost, string_of_int (!arg_idx))$>>
         | _  -> failwith "error ret"
       in
       <:expr< 
	 let open Kirc in 
	 let a = {
	   ml_kern = $gen_args$;
	   body = $gen_body2$;
	   ret_val = $ret$;
		 extensions = [ $match !extensions with 
				| [] -> <:expr<>>
				| t::[] -> t 
				| _ -> exSem_of_list  !extensions$];
	 }
	 in
	 a>>
      ]
    ];
  k_patt:
    [
      [l = patt -> l
      | "->"; "kern"; l=patt -> l  
      ]
    ];
  (* kexpr:
     [ ";"
     ]*)
  kernel_body:
    [
      [exprs =  (*LIST0*) sequence -> exprs
(*{t=TUnknown; e=Seqs exprs; loc=_loc}*)
      ]
    ];
  sequence':
    [ 
      [ 
	->fun e -> e
      | ";" -> fun e -> e
      | ";"; el = sequence -> fun e ->
      {t=TUnknown; e=Seq(_loc, e, el); loc = _loc}
      ] 
    ]
  ;
  sequence:
    [
      [ 
	e = kexpr; k = sequence' -> k e 
      ]
    ];
  do_sequence:
    [
      [seq = TRY ["{"; seq = sequence; "}" -> seq] -> seq
      | TRY ["{"; "}"] -> {t=TUnknown; e=Noop; loc=_loc}
      | seq = TRY [seq = sequence; "done" -> seq] -> seq
      | "done" -> {t=TUnknown; e=Noop; loc=_loc}
      ]
];
  kexpr:
    [ 
      ["let"; opt_mutable = OPT "mutable";  var = ident; "="; y = SELF; "in"; z = sequence  ->  
	{t=TUnknown; 
	 e=  Bind(_loc, 
                  {t= TUnknown; 
                   e= Id (_loc, var);
                   loc = _loc},
                  y, z, (match opt_mutable with
		  | None -> false
		  | _ -> true)
	 );
         loc = _loc}
      |"let"; "open"; i = module_longident; "in"; s = sequence ->
	{
	  t = TUnknown;
	  e = Open (_loc, i, s);
	  loc = _loc
	}]
    | "if"
	[ "if"; cond=SELF; "then"; cons1=sequence; 
	  "else"; cons2=sequence -> 
	  {t=TUnknown; e= Ife(_loc,cond,cons1,cons2); loc = _loc}
	| "if"; cond=SELF; "then"; cons1 = sequence -> 
	{t=TUnknown; e= If(_loc,cond,cons1); loc = _loc}
	]
    | "+" LEFTA
	[ x = SELF; "+!"; y = SELF -> {t=TInt32; e = Plus32(_loc, x,y); loc = _loc}
	| x = SELF; "+!!"; y = SELF -> {t=TInt64; e = Plus64(_loc, x,y); loc = _loc}
	| x = SELF; "+"; y = SELF -> {t=TInt; e = Plus(_loc, x,y); loc = _loc}
	| x = SELF; "+."; y = SELF -> {t=TFloat32; e = PlusF32(_loc, x,y); loc = _loc}]
    | "-" LEFTA
	[ x = SELF; "-!"; y = SELF -> {t=TInt32; e = Min32(_loc, x,y); loc = _loc}
	| x = SELF; "-!!"; y = SELF -> {t=TInt64; e = Min64(_loc, x,y); loc = _loc}
	| x = SELF; "-"; y = SELF -> {t=TInt; e = Min(_loc, x,y); loc = _loc}
	| x = SELF; "-."; y = SELF -> {t=TFloat32; e = MinF32(_loc, x,y); loc = _loc}]

    | "*" LEFTA
	[ x = SELF; "*!"; y = SELF -> {t=TInt32; e = Mul32(_loc, x,y); loc = _loc}
	| x = SELF; "*!!"; y = SELF -> {t=TInt64; e = Mul64(_loc, x,y); loc = _loc}
	| x = SELF; "*"; y = SELF -> {t=TInt; e = Mul(_loc, x,y); loc = _loc}
	| x = SELF; "*."; y = SELF -> {t=TFloat32; e = MulF32(_loc, x,y); loc = _loc}]
    | "mod" LEFTA
	[ x = SELF; "mod"; y = SELF -> {t=TInt; e = Mod(_loc, x,y); loc = _loc}]
    | "/" LEFTA
	[ x = SELF; "/!"; y = SELF -> {t=TInt32; e = Div32(_loc, x,y); loc = _loc}
	| x = SELF; "/!!"; y = SELF -> {t=TInt64; e = Div64(_loc, x,y); loc = _loc}
	| x = SELF; "/"; y = SELF -> {t=TInt; e = Div(_loc, x,y); loc = _loc}
	| x = SELF; "/."; y = SELF -> {t=TFloat32; e = DivF32(_loc, x,y); loc = _loc}]
    
    | "||" LEFTA	
	[x = SELF; "||"; y = SELF -> {t=TBool; 
				      e = BoolOr (_loc, x, y); loc = _loc} ]
    
    | "&&" RIGHTA	
	[x = SELF; "&&"; y = SELF -> {t=TBool; e = BoolAnd (_loc, x, y); loc = _loc} ]
   
		
	  | "apply" LEFTA
	[ e1 = SELF; e2 = SELF -> {t=(TUnknown); e= App(_loc, e1, [e2]); loc = _loc}
	]	
	 
    | "." RIGHTA
        [x = SELF; "."; "[<"; y=SELF; ">]"  -> {t=(TUnknown); 
						e = VecGet (_loc, x, y); loc = _loc}
		|l = UIDENT ; "."; e = SELF -> {t=(TUnknown); 
						 e = ModuleAccess (_loc, l, e); 
						 loc = _loc}
	]
    | "loop"
	[ "for"; x = ident; "="; y=SELF; "to"; z = SELF; "do";  body=do_sequence -> 
	{t = TUnit; e = DoLoop (_loc, 
				{t= TInt; 
				 e= Id (_loc, x);
				 loc = _loc}
				  ,y,z,body); loc = _loc}
	| "while"; cond = sequence; "do"; body = do_sequence -> 
	{t = TUnit; e = While (_loc,cond, body); loc = _loc}] 
	
    | "="
	[ x=SELF; "="; y=SELF -> {t=TBool; e= BoolEq(_loc,x,y); loc = _loc}
	| x=SELF; "=!"; y=SELF -> {t=TBool; e= BoolEq32(_loc,x,y); loc = _loc}
	| x=SELF; "=!!"; y=SELF -> {t=TBool; e= BoolEq64(_loc,x,y); loc = _loc}
	| x=SELF; "=."; y=SELF -> {t=TBool; e= BoolEqF(_loc,x,y); loc = _loc}]
    | "<"
	[ x=SELF; "<"; y=SELF -> {t=TBool; e= BoolLt(_loc,x,y); loc = _loc}
	| x=SELF; "<!"; y=SELF -> {t=TBool; e= BoolLt32(_loc,x,y); loc = _loc}
	| x=SELF; "<!!"; y=SELF -> {t=TBool; e= BoolLt64(_loc,x,y); loc = _loc}
	| x=SELF; "<."; y=SELF -> {t=TBool; e= BoolLtF(_loc,x,y); loc = _loc}]

    | "<="
	[ x=SELF; "<="; y=SELF -> {t=TBool; e= BoolLtE(_loc,x,y); loc = _loc}
	| x=SELF; "<=!"; y=SELF -> {t=TBool; e= BoolLtE32(_loc,x,y); loc = _loc}
	| x=SELF; "<=!!"; y=SELF -> {t=TBool; e= BoolLtE64(_loc,x,y); loc = _loc}
	| x=SELF; "<=."; y=SELF -> {t=TBool; e= BoolLtEF(_loc,x,y); loc = _loc}]

    | ">"
	[ x=SELF; ">"; y=SELF -> {t=TBool; e= BoolGt(_loc,x,y); loc = _loc}
	| x=SELF; ">!"; y=SELF -> {t=TBool; e= BoolGt32(_loc,x,y); loc = _loc}
	| x=SELF; ">!!"; y=SELF -> {t=TBool; e= BoolGt64(_loc,x,y); loc = _loc}
	| x=SELF; ">."; y=SELF -> {t=TBool; e= BoolGtF(_loc,x,y); loc = _loc}]

   | ">="
	[ x=SELF; ">="; y=SELF -> {t=TBool; e= BoolGtE(_loc,x,y); loc = _loc}
	| x=SELF; ">=!"; y=SELF -> {t=TBool; e= BoolGtE32(_loc,x,y); loc = _loc}
	| x=SELF; ">=!!"; y=SELF -> {t=TBool; e= BoolGtE64(_loc,x,y); loc = _loc}
	| x=SELF; ">=."; y=SELF -> {t=TBool; e= BoolGtEF(_loc,x,y); loc = _loc}]



    | "simple"
      	["(" ;  x= sequence; ")"  ->  x
	| "("; ")" -> {t=TUnit; e = Noop; loc = _loc}
	| x = FLOAT-> {t=TFloat; e = Float32 (_loc, x); loc = _loc}
(*	| "Float32"; "."; "("; x = sequence ; ")" -> 
	(try (float32_expr x) 
         with 
         | TypeError(expected, given, loc) -> 
           failwith ("Type Error : expecting : "^
			(ktyp_to_string expected)^" but given : "^
			(ktyp_to_string given)^" in position : "^(Loc.to_string loc)))] 
	|
	[ "Float64"; "."; "("; x = sequence ; ")" -> 
	(try (float64_expr x)
        with 
         | TypeError(expected, given, loc) -> 
           failwith ("Type Error : expecting : "^
			(ktyp_to_string expected)^" but given : "^
			(ktyp_to_string given)^" in position : "^(Loc.to_string loc))); *)
	|x = LIDENT  -> {t=TUnknown; e = Id (_loc, IdLid(_loc,x)); loc = _loc}
      	|x = INT32  ->{t=TInt32; e = Int32 (_loc, x); loc = _loc}
      	|x = INT  ->{t=TInt; e = Int (_loc, x); loc = _loc}] 				
    | "<-" 
	[ x = SELF; "<-"; y= SELF  -> {t=(TUnit); e = VecSet (_loc, x, y); loc = _loc}
	]
    | ":=" 
	[ x = SELF; ":="; y= SELF  -> {t=(TUnit); e = Acc (_loc, x, y); loc = _loc}
	]
    | "!"
    [ "!"; x = ident -> 
    {t=TUnknown; 
     e=Ref(_loc,
	   {t=TUnknown; e = Id (_loc, x); loc = _loc}
     ); loc = _loc}
    ]
    
    ];


(*  | [x = SELF; "EOI"  -> {t=TUnknown; e= End(_loc, x); loc = _loc}]*)
  

  END
